import logging
from threading import RLock
from typing import Mapping, List, Tuple, Union, Callable, Any, Optional

from multiprocessing.pool import ThreadPool

from errbot import Message
from errbot.backends.base import Identifier, Room, RoomOccupant

log = logging.getLogger(__name__)

Predicate = Callable[[Mapping[str, Any]], bool]

EXECUTOR_THREADS = 5  # the maximum number of simultaneous flows in automatic mode at the same time.


class FlowNode(object):
    """
    This is a step in a Flow/conversation. It is linked to a specific botcmd and also a "predicate".

    The predicate is a function that tells the flow executor if the flow can enter the step without the user
    intervention (automatically). The predicates defaults to False.

    The predicate is a function that takes one parameter, the context of the conversation.
    """

    def __init__(self, command: str = None, hints: bool = True):
        """
        Creates a FlowNone, takes the command to which the Node is linked to.
        :param command: the command this Node is linked to. Can only be None if this Node is a Root.
        :param hints: hints the users for the next steps in chat.
        """
        self.command = command
        self.children = []  # (predicate, node)
        self.hints = hints

    def connect(self, node_or_command: Union['FlowNode', str], predicate: Predicate = lambda _: False,
                hints: bool = True):
        """
        Construct the flow graph by connecting this node to another node or a command.
        The predicate is a function that tells the flow executor if the flow can enter the step without the user
        intervention (automatically).
        :param node_or_command: the node or a string for a command you want to connect this Node to
                               (this node or command will be the follow up of this one)
        :param predicate: function with one parameter, the context, to determine of the flow executor can continue
                           automatically this flow with no user intervention.
        :param hints: hints the user on the next step possible.
        :return: the newly created node if you passed a command or the node you gave it to be easily chainable.
        """
        node_to_connect_to = node_or_command if isinstance(node_or_command, FlowNode) else FlowNode(node_or_command,
                                                                                                    hints=hints)
        self.children.append((predicate, node_to_connect_to))
        return node_to_connect_to

    def predicate_for_node(self, node: 'FlowNode'):
        """
        gets the predicate function for the specified child node.
        :param node: the child node
        :return: the predicate that allows the automatic execution of that node.
        """
        for predicate, possible_node in self.children:
            if node == possible_node:
                return predicate
        return None

    def __str__(self):
        return self.command


class FlowRoot(FlowNode):
    """
    This represent the entry point of a flow description.
    """

    def __init__(self, name: str, description: str):
        """

        :param name: The name of the conversation/flow.
        :param description:  A human description of what this flow does.
        :param hints: Hints for the next steps when triggered.
        """
        super().__init__()
        self.name = name
        self.description = description
        self.auto_triggers = set()
        self.room_flow = False

    def connect(self,
                node_or_command: Union['FlowNode', str],
                predicate: Predicate = lambda _: False,
                auto_trigger: bool = False,
                room_flow: bool = False):
        """
        :see: FlowNode except fot auto_trigger
        :param predicate: :see: FlowNode
        :param node_or_command: :see: FlowNode
        :param auto_trigger: Flag this root as autotriggering: it will start a flow if this command is executed
                              in the chat.
        :param room_flow: Bind the flow to the room instead of a single person
        """
        resp = super().connect(node_or_command, predicate)
        if auto_trigger:
            self.auto_triggers.add(node_or_command)
        self.room_flow = room_flow
        return resp

    def __str__(self):
        return self.name


class _FlowEnd(FlowNode):
    def __str__(self):
        return 'END'


#: Flow marker indicating that the flow ends.
FLOW_END = _FlowEnd()


class InvalidState(Exception):
    """
    Raised when the Flow Executor is asked to do something contrary to the contraints it has been given.
    """
    pass


class Flow(object):
    """
    This is a live Flow. It keeps context of the conversation (requestor and context).
    Context is just a python dictionary representing the state of the conversation.
    """

    def __init__(self, root: FlowRoot, requestor: Identifier, initial_context: Mapping[str, Any]):
        """

        :param root: the root of this flow.
        :param requestor: the user requesting this flow.
        :param initial_context: any data we already have that could help executing this flow automatically.
        """
        self._root = root
        self._current_step = self._root
        self.ctx = dict(initial_context)
        self.requestor = requestor

    def next_autosteps(self) -> List[FlowNode]:
        """
        Get the next steps that can be automatically executed according to the set predicates.
        """
        return [node for predicate, node in self._current_step.children if predicate(self.ctx)]

    def next_steps(self) -> List[FlowNode]:
        """
        Get all the possible next steps after this one (predicates statisfied or not).
        """
        return [node for predicate, node in self._current_step.children]

    def advance(self, next_step: FlowNode, enforce_predicate=True):
        """
        Move on along the flow.
        :param next_step: Which node you want to move the flow forward to.
        :param enforce_predicate: Do you want to check if the predicate is verified for this step or not.
                                   Usually, if it is a manual step, the predicate is irrelevant because the user
                                   will give the missing information as parameters to the command.
        """
        if enforce_predicate:
            predicate = self._current_step.predicate_for_node(next_step)
            if predicate is None:
                raise ValueError(f'There is no such children: {next_step}.')

            if not predicate(self.ctx):
                raise InvalidState('It is not possible to advance to this step because its predicate is false.')

        self._current_step = next_step

    @property
    def name(self) -> str:
        """
        Helper property to get the name of the flow.
        """
        return self._root.name

    @property
    def current_step(self) -> FlowNode:
        """
        The current step this Flow is waiting on.
        """
        return self._current_step

    @property
    def root(self) -> FlowRoot:
        """
        The original flowroot of this flow.
        """
        return self._root

    def check_identifier(self, identifier: Identifier):
        is_room = isinstance(self.requestor, Room)
        is_room = is_room and isinstance(identifier, RoomOccupant)
        is_room = is_room and self.requestor == identifier.room
        return is_room or self.requestor == identifier

    def __str__(self):
        return f'{self._root} ({self.requestor}) with params {dict(self.ctx)}'


class BotFlow:
    """
    Defines a Flow plugin ie. a plugin that will define new flows from its methods with the @botflow decorator.
    """

    def __init__(self, bot, name=None):
        super().__init__()
        self._bot = bot
        self.is_activated = False
        self._name = name

    @property
    def name(self) -> str:
        """
        Get the name of this flow as described in its .plug file.

        :return: The flow name.
        """
        return self._name

    def activate(self) -> None:
        """
            Override if you want to do something at initialization phase (don't forget to
            super(Gnagna, self).activate())
        """
        self._bot.inject_flows_from(self)
        self.is_activated = True

    def deactivate(self) -> None:
        """
            Override if you want to do something at tear down phase (don't forget to super(Gnagna, self).deactivate())
        """
        self._bot.remove_flows_from(self)
        self.is_activated = False

    def get_command(self, command_name: str):
        """
            Helper to get a specific command.
        """
        self._bot.all_commands.get(command_name, None)


class FlowExecutor(object):
    """
    This is a instance that can monitor and execute flow instances.
    """

    def __init__(self, bot):
        self._lock = RLock()
        self.flow_roots = {}
        self.in_flight = []
        self._pool = ThreadPool(EXECUTOR_THREADS)
        self._bot = bot

    def add_flow(self, flow: FlowRoot):
        """
        Register a flow with this executor.
        """
        with self._lock:
            self.flow_roots[flow.name] = flow

    def trigger(self, cmd: str, requestor: Identifier, extra_context=None) -> Optional[Flow]:
        """
        Trigger workflows that may have command cmd as a auto_trigger or an in flight flow waiting for command.
        This assume cmd has been correctly executed.
        :param requestor: the identifier of the person who started this flow
        :param cmd: the command that has just been executed.
        :param extra_context: extra context from the current conversation
        :returns: The flow it triggered or None if none were matching.
        """
        flow, next_step = self.check_inflight_flow_triggered(cmd, requestor)
        if not flow:
            flow, next_step = self._check_if_new_flow_is_triggered(cmd, requestor)
        if not flow:
            return None

        flow.advance(next_step, enforce_predicate=False)
        if extra_context:
            flow.ctx = dict(extra_context)
        self._enqueue_flow(flow)
        return flow

    def check_inflight_already_running(self, user: Identifier) -> bool:
        """
            Check if user is already running a flow.
        :param user: the user
        """
        with self._lock:
            for flow in self.in_flight:
                if flow.requestor == user:
                    return True
        return False

    def check_inflight_flow_triggered(self, cmd: str, user: Identifier) -> Tuple[Optional[Flow], Optional[FlowNode]]:
        """
        Check if a command from a specific user was expected in one of the running flow.
        :param cmd: the command that has just been executed.
        :param user: the identifier of the person who started this flow
        :returns: The name of the flow it triggered or None if none were matching."""
        log.debug("Test if the command %s is a trigger for an inflight flow ...", cmd)
        # TODO: What if 2 flows wait for the same command ?
        with self._lock:
            for flow in self.in_flight:
                if flow.check_identifier(user):
                    log.debug("Requestor has a flow %s in flight", flow.name)
                    for next_step in flow.next_steps():
                        if next_step.command == cmd:
                            log.debug("Requestor has a flow in flight waiting for this command !")
                            return flow, next_step
        log.debug("None matched.")
        return None, None

    def _check_if_new_flow_is_triggered(self, cmd: str, user: Identifier) -> Tuple[Optional[Flow], Optional[FlowNode]]:
        """
        Trigger workflows that may have command cmd as a auto_trigger..
        This assume cmd has been correctly executed.
        :param cmd: the command that has just been executed.
        :param user: the identifier of the person who started this flow
        :returns: The name of the flow it triggered or None if none were matching.
        """
        log.debug("Test if the command %s is an auto-trigger for any flow ...", cmd)
        with self._lock:
            for name, flow_root in self.flow_roots.items():
                if cmd in flow_root.auto_triggers and not self.check_inflight_already_running(user):
                    log.debug("Flow %s has been auto-triggered by the command %s by user %s", name, cmd, user)
                    return self._create_new_flow(flow_root, user, cmd)
        return None, None

    @staticmethod
    def _create_new_flow(flow_root, requestor: Identifier, initial_command) \
            -> Tuple[Optional[Flow], Optional[FlowNode]]:
        """
        Helper method to create a new FLow.
        """
        empty_context = {}
        flow = Flow(flow_root, requestor, empty_context)
        for possible_next_step in flow.next_steps():
            if possible_next_step.command == initial_command:
                # The predicate is good as we just executed manually the command.
                return flow, possible_next_step
        return None, None

    def start_flow(self, name: str, requestor: Identifier, initial_context: Mapping[str, Any]) -> Flow:
        """
        Starts the execution of a Flow.
        """
        if name not in self.flow_roots:
            raise ValueError(f'Flow {name} doesn\'t exist')
        if self.check_inflight_already_running(requestor):
            raise ValueError(f'User {str(requestor)} is already running a flow.')

        flow_root = self.flow_roots[name]
        identity = requestor
        if isinstance(requestor, RoomOccupant) and flow_root.room_flow:
            identity = requestor.room

        flow = Flow(self.flow_roots[name], identity, initial_context)
        self._enqueue_flow(flow)
        return flow

    def stop_flow(self, name: str, requestor: Identifier) -> Optional[Flow]:
        """
        Stops a specific flow. It is a no op if the flow doesn't exist.
        Returns the stopped flow if found.
        """
        with self._lock:
            for flow in self.in_flight:
                if flow.name == name and flow.check_identifier(requestor):
                    log.debug(f'Removing flow {str(flow)}.')
                    self.in_flight.remove(flow)
                    return flow
        return None

    def _enqueue_flow(self, flow):
        with self._lock:
            if flow not in self.in_flight:
                self.in_flight.append(flow)
        self._pool.apply_async(self.execute, (flow,))

    def execute(self, flow: Flow):
        """
        This is where the flow execution happens from one of the thread of the pool.
        """
        while True:
            autosteps = flow.next_autosteps()
            steps = flow.next_steps()

            if not steps:
                log.debug("Flow ended correctly.Nothing left to do.")
                with self._lock:
                    self.in_flight.remove(flow)
                break

            if not autosteps and flow.current_step.hints:
                possible_next_steps = [f'You are in the flow **{flow.name}**, you can continue with:\n\n']
                for step in steps:
                    cmd = step.command
                    cmd_fnc = self._bot.all_commands[cmd]
                    reg_cmd = cmd_fnc._err_re_command
                    syntax_args = cmd_fnc._err_command_syntax
                    reg_prefixed = cmd_fnc._err_command_prefix_required if reg_cmd else True
                    syntax = self._bot.prefix if reg_prefixed else ''
                    if not reg_cmd:
                        syntax += cmd.replace('_', ' ')
                    if syntax_args:
                        syntax += syntax_args
                    possible_next_steps.append(f'- {syntax}')
                self._bot.send(flow.requestor, '\n'.join(possible_next_steps))
                break

            log.debug('Steps triggered automatically %s.', ', '.join(str(node) for node in autosteps))
            log.debug('All possible next steps: %s.', ', '.join(str(node) for node in steps))

            for autostep in autosteps:
                log.debug("Proceeding automatically with step %s", autostep)
                if autostep == FLOW_END:
                    log.debug('This flow ENDED.')
                    with self._lock:
                        self.in_flight.remove(flow)
                    return
                try:
                    msg = Message(frm=flow.requestor, flow=flow)
                    result = self._bot.commands[autostep.command](msg, None)
                    log.debug('Step result %s: %s', flow.requestor, result)

                except Exception as e:
                    log.exception('%s errored at %s', flow, autostep)
                    self._bot.send(flow.requestor, f'{flow} errored at {autostep} with "{e}"')
                flow.advance(autostep)  # TODO: this is only true for a single step, make it forkable.
        log.debug('Flow execution suspended/ended normally.')



import logging
import shlex
from threading import Timer, current_thread
from types import ModuleType
from typing import Tuple, Callable, Mapping, Sequence
from io import IOBase
import re

from .storage import StoreMixin, StoreNotOpenError
from errbot.backends.base import Message, Presence, Stream, Room, Identifier, ONLINE, Card

log = logging.getLogger(__name__)


class ValidationException(Exception):
    pass


def recurse_check_structure(sample, to_check):
    sample_type = type(sample)
    to_check_type = type(to_check)

    # Skip this check if the sample is None because it will always be something
    # other than NoneType when changed from the default. Raising ValidationException
    # would make no sense then because it would defeat the whole purpose of having
    # that key in the sample when it could only ever be None.
    if sample is not None and sample_type != to_check_type:
        raise ValidationException(f'{sample} [{sample_type}] is not the same type as {to_check} [{to_check_type}].')

    if sample_type in (list, tuple):
        for element in to_check:
            recurse_check_structure(sample[0], element)
        return

    if sample_type == dict:
        for key in sample:
            if key not in to_check:
                raise ValidationException(f'{to_check} doesn\'t contain the key {key}.')
        for key in to_check:
            if key not in sample:
                raise ValidationException(f'{to_check} contains an unknown key {key}.')
        for key in sample:
            recurse_check_structure(sample[key], to_check[key])
        return


class CommandError(Exception):
    """
    Use this class to report an error condition from your commands, the command
    did not proceed for a known "business" reason.
    """

    def __init__(self, reason: str, template: str = None):
        """
        :param reason: the reason for the error in the command.
        :param template: apply this specific template to report the error.
        """
        self.reason = reason
        self.template = template

    def __str__(self):
        return str(self.reason)


class Command(object):
    """
    This is a dynamic definition of an errbot command.
    """

    def __init__(self, function, cmd_type=None, cmd_args=None, cmd_kwargs=None, name=None, doc=None):
        """
        Create a Command definition.

        :param function:
            a function or a lambda with the correct signature for the type of command to inject for example `def
            mycmd(plugin, msg, args)` for a botcmd.  Note: the first parameter will be the plugin itself (equivalent to
            self).
        :param cmd_type:
            defaults to `botcmd` but can be any decorator function used for errbot commands.
        :param cmd_args: the parameters of the decorator.
        :param cmd_kwargs: the kwargs parameter of the decorator.
        :param name:
            defaults to the name of the function you are passing if it is a first class function or needs to be set if
            you use a lambda.
        :param doc:
            defaults to the doc of the given function if it is a first class function. It can be set for a lambda or
            overridden for a function with this.  """
        if cmd_type is None:
            from errbot import botcmd  # TODO refactor this out of __init__ so it can be reusable.
            cmd_type = botcmd
        if name is None:
            if function.__name__ == '<lambda>':
                raise ValueError('function is a lambda (anonymous), parameter name needs to be set.')
            name = function.__name__
        self.name = name
        if cmd_kwargs is None:
            cmd_kwargs = {}
        if cmd_args is None:
            cmd_args = ()
        function.__name__ = name
        if doc:
            function.__doc__ = doc
        self.definition = cmd_type(*((function,) + cmd_args), **cmd_kwargs)


# noinspection PyAbstractClass
class BotPluginBase(StoreMixin):
    """
     This class handle the basic needs of bot plugins like loading, unloading and creating a storage
     It is the main contract between the plugins and the bot
    """

    def __init__(self, bot, name=None):
        self.is_activated = False
        self.current_pollers = []
        self.current_timers = []
        self.dependencies = []
        self._dynamic_plugins = {}
        self.log = logging.getLogger(f'errbot.plugins.{name}')
        self.log.debug('Logger for plugin initialized...')
        self._bot = bot
        self.plugin_dir = bot.repo_manager.plugin_dir
        self._name = name
        super().__init__()

    @property
    def name(self) -> str:
        """
        Get the name of this plugin as described in its .plug file.

        :return: The plugin name.
        """
        return self._name

    @property
    def mode(self) -> str:
        """
        Get the current active backend.

        :return: the mode like 'tox', 'xmpp' etc...
        """
        return self._bot.mode

    @property
    def bot_config(self) -> ModuleType:
        """
        Get the bot configuration from config.py.
        For example you can access:
        self.bot_config.BOT_DATA_DIR
        """
        # if BOT_ADMINS is just an unique string make it a tuple for backwards
        # compatibility
        if isinstance(self._bot.bot_config.BOT_ADMINS, str):
            self._bot.bot_config.BOT_ADMINS = (self._bot.bot_config.BOT_ADMINS,)
        return self._bot.bot_config

    @property
    def bot_identifier(self) -> Identifier:
        """
        Get bot identifier on current active backend.

        :return Identifier
        """
        return self._bot.bot_identifier

    def init_storage(self) -> None:
        log.debug(f'Init storage for {self.name}.')
        self.open_storage(self._bot.storage_plugin, self.name)

    def activate(self) -> None:
        """
            Override if you want to do something at initialization phase (don't forget to
            super(Gnagna, self).activate())
        """
        self.init_storage()
        self._bot.inject_commands_from(self)
        self._bot.inject_command_filters_from(self)
        self.is_activated = True

    def deactivate(self) -> None:
        """
            Override if you want to do something at tear down phase (don't forget to super(Gnagna, self).deactivate())
        """
        if self.current_pollers:
            log.debug('You still have active pollers at deactivation stage, I cleaned them up for you.')
            self.current_pollers = []
            for timer in self.current_timers:
                timer.cancel()

        try:
            self.close_storage()
        except StoreNotOpenError:
            pass
        self._bot.remove_command_filters_from(self)
        self._bot.remove_commands_from(self)
        self.is_activated = False

        for plugin in self._dynamic_plugins.values():
            self._bot.remove_command_filters_from(plugin)
            self._bot.remove_commands_from(plugin)

    def start_poller(self,
                     interval: float,
                     method: Callable[..., None],
                     times: int = None,
                     args: Tuple = None,
                     kwargs: Mapping = None):
        """ Starts a poller that will be called at a regular interval

        :param interval: interval in seconds
        :param method: targetted method
        :param times:
            number of times polling should happen (defaults to``None`` which
            causes the polling to happen indefinitely)
        :param args: args for the targetted method
        :param kwargs: kwargs for the targetting method
        """
        if not kwargs:
            kwargs = {}
        if not args:
            args = []

        log.debug(f'Programming the polling of {method.__name__} every {interval} seconds '
                  f'with args {str(args)} and kwargs {str(kwargs)}')
        # noinspection PyBroadException
        try:
            self.current_pollers.append((method, args, kwargs))
            self.program_next_poll(interval, method, times, args, kwargs)
        except Exception:
            log.exception('Poller programming failed.')

    def stop_poller(self,
                    method: Callable[..., None],
                    args: Tuple = None,
                    kwargs: Mapping = None):
        if not kwargs:
            kwargs = {}
        if not args:
            args = []
        log.debug(f'Stop polling of {method} with args {args} and kwargs {kwargs}')
        self.current_pollers.remove((method, args, kwargs))

    def program_next_poll(self,
                          interval: float,
                          method: Callable[..., None],
                          times: int = None,
                          args: Tuple = None,
                          kwargs: Mapping = None):
        if times is not None and times <= 0:
            return

        t = Timer(interval=interval, function=self.poller,
                  kwargs={'interval': interval, 'method': method,
                          'times': times, 'args': args, 'kwargs': kwargs})
        self.current_timers.append(t)  # save the timer to be able to kill it
        t.setName(f'Poller thread for {type(method.__self__).__name__}')
        t.setDaemon(True)  # so it is not locking on exit
        t.start()

    def poller(self,
               interval: float,
               method: Callable[..., None],
               times: int = None,
               args: Tuple = None,
               kwargs: Mapping = None):
        previous_timer = current_thread()
        if previous_timer in self.current_timers:
            log.debug('Previous timer found and removed')
            self.current_timers.remove(previous_timer)

        if (method, args, kwargs) in self.current_pollers:
            # noinspection PyBroadException
            try:
                method(*args, **kwargs)
            except Exception:
                log.exception('A poller crashed')

            if times is not None:
                times -= 1

            self.program_next_poll(interval, method, times, args, kwargs)

    def create_dynamic_plugin(self, name: str, commands: Tuple[Command], doc: str = ''):
        """
            Creates a plugin dynamically and exposes its commands right away.

            :param name: name of the plugin.
            :param commands: a tuple of command definition.
            :param doc: the main documentation of the plugin.
        """
        if name in self._dynamic_plugins:
            raise ValueError('Dynamic plugin %s already created.')
        # cleans the name to be a valid python type.
        plugin_class = type(re.sub(r'\W|^(?=\d)', '_', name), (BotPlugin,),
                            {command.name: command.definition for command in commands})
        plugin_class.__errdoc__ = doc
        plugin = plugin_class(self._bot, name=name)
        self._dynamic_plugins[name] = plugin
        self._bot.inject_commands_from(plugin)

    def destroy_dynamic_plugin(self, name: str):
        """
            Reverse operation of create_dynamic_plugin.

            This allows you to dynamically refresh the list of commands for example.
            :param name: the name of the dynamic plugin given to create_dynamic_plugin.
        """
        if name not in self._dynamic_plugins:
            raise ValueError("Dynamic plugin %s doesn't exist.", name)
        plugin = self._dynamic_plugins[name]
        self._bot.remove_command_filters_from(plugin)
        self._bot.remove_commands_from(plugin)
        del self._dynamic_plugins[name]

    def get_plugin(self, name) -> 'BotPlugin':
        """
        Gets a plugin your plugin depends on. The name of the dependency needs to be listed in [Code] section
        key DependsOn of your plug file. This method can only be used after your plugin activation
        (or having called super().activate() from activate itself).
        It will return a plugin object.

        :param name: the name
        :return: the BotPlugin object requested.
        """
        if not self.is_activated:
            raise Exception('Plugin needs to be in activated state to be able to get its dependencies.')

        if name not in self.dependencies:
            raise Exception(f'Plugin dependency {name} needs to be listed in section [Core] key '
                            f'"DependsOn" to be used in get_plugin.')

        return self._bot.plugin_manager.get_plugin_obj_by_name(name)


# noinspection PyAbstractClass
class BotPlugin(BotPluginBase):

    def get_configuration_template(self) -> Mapping:
        """
        If your plugin needs a configuration, override this method and return
        a configuration template.

        For example a dictionary like:
        return {'LOGIN' : 'example@example.com', 'PASSWORD' : 'password'}

        Note: if this method returns None, the plugin won't be configured
        """
        return None

    def check_configuration(self, configuration: Mapping) -> None:
        """
        By default, this method will do only a BASIC check. You need to override
        it if you want to do more complex checks. It will be called before the
        configure callback. Note if the config_template is None, it will never
        be called.

        It means recusively:

        1. in case of a dictionary, it will check if all the entries and from
           the same type are there and not more.
        2. in case of an array or tuple, it will assume array members of the
           same type of first element of the template (no mix typed is supported)

        In case of validation error it should raise a errbot.ValidationException

        :param configuration: the configuration to be checked.
        """
        recurse_check_structure(self.get_configuration_template(), configuration)  # default behavior

    def configure(self, configuration: Mapping) -> None:
        """
        By default, it will just store the current configuration in the self.config
        field of your plugin. If this plugin has no configuration yet, the framework
        will call this function anyway with None.

        This method will be called before activation so don't expect to be activated
        at that point.

        :param configuration: injected configuration for the plugin.
        """
        self.config = configuration

    def activate(self) -> None:
        """
            Triggered on plugin activation.

            Override this method if you want to do something at initialization phase
            (don't forget to `super().activate()`).
        """
        super().activate()

    def deactivate(self) -> None:
        """
            Triggered on plugin deactivation.

            Override this method if you want to do something at tear-down phase
            (don't forget to `super().deactivate()`).
        """
        super().deactivate()

    def callback_connect(self) -> None:
        """
            Triggered when the bot has successfully connected to the chat network.

            Override this method to get notified when the bot is connected.
        """
        pass

    def callback_message(self, message: Message) -> None:
        """
            Triggered on every message not coming from the bot itself.

            Override this method to get notified on *ANY* message.

            :param message:
                representing the message that was received.
        """
        pass

    def callback_mention(self, message: Message, mentioned_people: Sequence[Identifier]) -> None:
        """
            Triggered if there are mentioned people in message.

            Override this method to get notified when someone was mentioned in message.
            [Note: This might not be implemented by all backends.]

            :param message:
                representing the message that was received.
            :param mentioned_people:
                all mentioned people in this message.
        """
        pass

    def callback_presence(self, presence: Presence) -> None:
        """
            Triggered on every presence change.

            :param presence:
                An instance of :class:`~errbot.backends.base.Presence`
                representing the new presence state that was received.
        """
        pass

    def callback_stream(self, stream: Stream) -> None:
        """
            Triggered asynchronously (in a different thread context) on every incoming stream
            request or file transfer request.
            You can block this call until you are done with the stream.
            To signal that you accept / reject the file, simply call stream.accept()
            or stream.reject() and return.

            :param stream:
                the incoming stream request.
        """
        stream.reject()  # by default, reject the file as the plugin doesn't want it.

    def callback_botmessage(self, message: Message):
        """
            Triggered on every message coming from the bot itself.

            Override this method to get notified on all messages coming from
            the bot itself (including those from other plugins).

            :param message:
                An instance of :class:`~errbot.backends.base.Message`
                representing the message that was received.
        """
        pass

    def callback_room_joined(self, room: Room):
        """
            Triggered when the bot has joined a MUC.

            :param room:
                An instance of :class:`~errbot.backends.base.MUCRoom`
                representing the room that was joined.
        """
        pass

    def callback_room_left(self, room: Room):
        """
            Triggered when the bot has left a MUC.

            :param room:
                An instance of :class:`~errbot.backends.base.MUCRoom`
                representing the room that was left.
        """
        pass

    def callback_room_topic(self, room: Room):
        """
            Triggered when the topic in a MUC changes.

            :param room:
                An instance of :class:`~errbot.backends.base.MUCRoom`
                representing the room for which the topic changed.
        """
        pass

    # Proxyfy some useful tools from the motherbot
    # this is basically the contract between the plugins and the main bot

    def warn_admins(self, warning: str) -> None:
        """
        Send a warning to the administrators of the bot.

        :param warning: The markdown-formatted text of the message to send.
        """
        self._bot.warn_admins(warning)

    def send(self,
             identifier: Identifier,
             text: str,
             in_reply_to: Message = None,
             groupchat_nick_reply: bool = False) -> None:
        """
            Send a message to a room or a user.

            :param groupchat_nick_reply: if True the message will mention the user in the chatroom.
            :param in_reply_to: the original message this message is a reply to (optional).
                                In some backends it will start a thread.
            :param text: markdown formatted text to send to the user.
            :param identifier: An Identifier representing the user or room to message.
                               Identifiers may be created with :func:`build_identifier`.
        """
        if not isinstance(identifier, Identifier):
            raise ValueError("identifier needs to be of type Identifier, the old string behavior is not supported")
        return self._bot.send(identifier, text, in_reply_to, groupchat_nick_reply)

    def send_card(self,
                  body: str = '',
                  to: Identifier = None,
                  in_reply_to: Message = None,
                  summary: str = None,
                  title: str = '',
                  link: str = None,
                  image: str = None,
                  thumbnail: str = None,
                  color: str = 'green',
                  fields: Tuple[Tuple[str, str], ...] = ()) -> None:
        """
        Sends a card.

        A Card is a special type of preformatted message. If it matches with a backend similar concept like on
        Slack or Hipchat it will be rendered natively, otherwise it will be sent as a regular formatted message.

        :param body: main text of the card in markdown.
        :param to: the card is sent to this identifier (Room, RoomOccupant, Person...).
        :param in_reply_to: the original message this message is a reply to (optional).
        :param summary: (optional) One liner summary of the card, possibly collapsed to it.
        :param title: (optional) Title possibly linking.
        :param link: (optional) url the title link is pointing to.
        :param image: (optional) link to the main image of the card.
        :param thumbnail: (optional) link to an icon / thumbnail.
        :param color: (optional) background color or color indicator.
        :param fields: (optional) a tuple of (key, value) pairs.
        """
        frm = in_reply_to.to if in_reply_to else self.bot_identifier
        if to is None:
            if in_reply_to is None:
                raise ValueError('Either to or in_reply_to needs to be set.')
            to = in_reply_to.frm
        self._bot.send_card(Card(body, frm, to, in_reply_to, summary, title, link, image, thumbnail, color, fields))

    def change_presence(self, status: str = ONLINE, message: str = '') -> None:
        """
            Changes the presence/status of the bot.

        :param status: One of the constant defined in base.py : ONLINE, OFFLINE, DND,...
        :param message: Additional message
        :return: None
        """
        self._bot.change_presence(status, message)

    def send_templated(self,
                       identifier: Identifier,
                       template_name: str,
                       template_parameters: Mapping,
                       in_reply_to: Message = None,
                       groupchat_nick_reply: bool = False) -> None:
        """
        Sends asynchronously a message to a room or a user.

        Same as send but passing a template name and parameters instead of directly the markdown text.
        :param template_parameters: arguments for the template.
        :param template_name: name of the template to use.
        :param groupchat_nick_reply: if True it will mention the user in the chatroom.
        :param in_reply_to: optionally, the original message this message is the answer to.
        :param identifier: identifier of the user or room to which you want to send a message to.
        """
        return self._bot.send_templated(identifier=identifier,
                                        template_name=template_name,
                                        template_parameters=template_parameters,
                                        in_reply_to=in_reply_to,
                                        groupchat_nick_reply=groupchat_nick_reply)

    def build_identifier(self, txtrep: str) -> Identifier:
        """
           Transform a textual representation of a user identifier to the correct
           Identifier object you can set in Message.to and Message.frm.

           :param txtrep: the textual representation of the identifier (it is backend dependent).
           :return: a user identifier.
        """
        return self._bot.build_identifier(txtrep)

    def send_stream_request(self,
                            user: Identifier,
                            fsource: IOBase,
                            name: str = None,
                            size: int = None,
                            stream_type: str = None):
        """
            Sends asynchronously a stream/file to a user.

            :param user: is the identifier of the person you want to send it to.
            :param fsource: is a file object you want to send.
            :param name: is an optional filename for it.
            :param size: is optional and is the espected size for it.
            :param stream_type: is optional for the mime_type of the content.

            It will return a Stream object on which you can monitor the progress of it.
        """
        return self._bot.send_stream_request(user, fsource, name, size, stream_type)

    def rooms(self) -> Sequence[Room]:
        """
        The list of rooms the bot is currently in.
        """
        return self._bot.rooms()

    def query_room(self, room: str) -> Room:
        """
        Query a room for information.

        :param room:
            The JID/identifier of the room to query for.
        :returns:
            An instance of :class:`~errbot.backends.base.MUCRoom`.
        :raises:
            :class:`~errbot.backends.base.RoomDoesNotExistError` if the room doesn't exist.
        """
        return self._bot.query_room(room)

    def start_poller(self,
                     interval: float,
                     method: Callable[..., None],
                     times: int = None,
                     args: Tuple = None,
                     kwargs: Mapping = None):
        """
            Start to poll a method at specific interval in seconds.

            Note: it will call the method with the initial interval delay for
            the first time

            Also, you can program
            for example : self.program_poller(self, 30, fetch_stuff)
            where you have def fetch_stuff(self) in your plugin

            :param interval: interval in seconds
            :param method: targetted method
            :param times:
                number of times polling should happen (defaults to``None``
                which causes the polling to happen indefinitely)
            :param args: args for the targetted method
            :param kwargs: kwargs for the targetting method

        """
        super().start_poller(interval, method, times, args, kwargs)

    def stop_poller(self,
                    method: Callable[..., None],
                    args: Tuple = None,
                    kwargs: Mapping = None):
        """
            stop poller(s).

            If the method equals None -> it stops all the pollers you need to
            regive the same parameters as the original start_poller to match a
            specific poller to stop

            :param kwargs: The initial kwargs you gave to start_poller.
            :param args: The initial args you gave to start_poller.
            :param method: The initial method you passed to start_poller.

        """
        super().stop_poller(method, args, kwargs)


class ArgParserBase(object):
    """
    The `ArgSplitterBase` class defines the API which is used for argument
    splitting (used by the `split_args_with` parameter on
    :func:`~errbot.decorators.botcmd`).
    """

    def parse_args(self, args: str):
        """
        This method takes a string of un-split arguments and parses it,
        returning a list that is the result of splitting.

        If splitting fails for any reason it should return an exception
        of some kind.

        :param args: string to parse
        """
        raise NotImplementedError()


class SeparatorArgParser(ArgParserBase):
    """
    This argument splitter splits args on a given separator, like
    :func:`str.split` does.
    """

    def __init__(self, separator: str = None, maxsplit: int = -1):
        """
        :param separator:
            The separator on which arguments should be split. If sep is
            None, any whitespace string is a separator and empty strings
            are removed from the result.
        :param maxsplit:
            If given, do at most this many splits.
        """
        self.separator = separator
        self.maxsplit = maxsplit

    def parse_args(self, args: str):
        return args.split(self.separator, self.maxsplit)


class ShlexArgParser(ArgParserBase):
    """
    This argument splitter splits args using posix shell quoting rules,
    like :func:`shlex.split` does.
    """

    def parse_args(self, args):
        return shlex.split(args)



from errbot import BotPlugin, botcmd

import borken  # fails on purpose


class Broken(BotPlugin):
    @botcmd
    def hello(self, msg, args):
        """ this command says hello """
        return 'Hello World !'



from __future__ import absolute_import
from errbot import BotPlugin, botcmd


class PollerPlugin(BotPlugin):

    def delayed_hello(self, frm):
        self.send(frm, 'Hello world! was sent 5 seconds ago')

    @botcmd
    def hello(self, msg, args):
        """Say hello to the world."""
        self.start_poller(0.1, self.delayed_hello, times=1,
                          kwargs={'frm': msg.frm})
        return "Hello, world!"

    def delayed_hello_loop(self, frm):
        self.send(frm, 'Hello world! was sent 5 seconds ago')
        self.stop_poller(self.delayed_hello_loop, args=(frm, ))

    @botcmd
    def hello_loop(self, msg, args):
        """Say hello to the world."""
        self.start_poller(0.1, self.delayed_hello_loop,
                          args=(msg.frm, ))
        return "Hello, world!"



"""Test _admins_to_notify wrapper functionality"""
import pytest


extra_config = {'BOT_ADMINS_NOTIFICATIONS': 'zoni@localdomain'}


def test_admins_to_notify(testbot):
    """Test which admins will be notified"""
    notified_admins = testbot._bot._admins_to_notify()
    assert 'zoni@localdomain' in notified_admins


def test_admins_not_notified(testbot):
    """Test which admins will not be notified"""
    notified_admins = testbot._bot._admins_to_notify()
    assert 'gbin@local' not in notified_admins



# config for travisci
# Don't use this for sensible defaults
import logging
BOT_DATA_DIR = '/tmp'
BOT_EXTRA_PLUGIN_DIR = None
AUTOINSTALL_DEPS = True
BOT_LOG_FILE = '/tmp/err.log'
BOT_LOG_LEVEL = logging.DEBUG
BOT_LOG_SENTRY = False
SENTRY_DSN = ''
SENTRY_LOGLEVEL = BOT_LOG_LEVEL
BOT_ASYNC = True
BOT_IDENTITY = {
    'username': 'err@localhost',
    'password': 'changeme',
}

BOT_ADMINS = ('gbin@localhost',)
CHATROOM_PRESENCE = ()
CHATROOM_FN = 'Err'
BOT_PREFIX = '!'
DIVERT_TO_PRIVATE = ()
CHATROOM_RELAY = {}
REVERSE_CHATROOM_RELAY = {}



import json
import logging
import os

import pytest
import requests
import socket
from errbot.backends.test import FullStackTest, testbot
from time import sleep

log = logging.getLogger(__name__)

PYTHONOBJECT = ['foo', {'bar': ('baz', None, 1.0, 2)}]
JSONOBJECT = json.dumps(PYTHONOBJECT)
# Webserver port is picked based on the process ID so that when tests
# are run in parallel with pytest-xdist, each process runs the server
# on a different port
WEBSERVER_PORT = 5000 + (os.getpid() % 1000)
WEBSERVER_SSL_PORT = WEBSERVER_PORT + 1000


def webserver_ready(host, port):
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    try:
        s.connect((host, port))
        s.shutdown(socket.SHUT_RDWR)
        s.close()
        return True
    except Exception:
        return False


extra_plugin_dir = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'webhooks_plugin')


def wait_for_server(port: int):
    failure_count = 10
    while not webserver_ready('localhost', port):
        waiting_time = 1.0 / failure_count
        log.info('Webserver not ready yet, sleeping for %f second.', waiting_time)
        sleep(waiting_time)
        failure_count -= 1
        if failure_count == 0:
            raise TimeoutError("Could not start the internal Webserver to test.")


@pytest.fixture
def webhook_testbot(request, testbot):
    testbot.push_message("!plugin config Webserver {'HOST': 'localhost', 'PORT': %s, 'SSL': None}" % WEBSERVER_PORT)
    log.info(testbot.pop_message())
    wait_for_server(WEBSERVER_PORT)
    return testbot


def test_not_configured_url_returns_404(webhook_testbot):
    assert requests.post(
        'http://localhost:{}/randomness_blah'.format(WEBSERVER_PORT),
        "{'toto': 'titui'}"
    ).status_code == 404


def test_webserver_plugin_ok(webhook_testbot):
    assert "/echo" in webhook_testbot.exec_command("!webstatus")


def test_trailing_no_slash_ok(webhook_testbot):
    assert requests.post(
        'http://localhost:{}/echo'.format(WEBSERVER_PORT),
        JSONOBJECT
    ).text == repr(json.loads(JSONOBJECT))


def test_trailing_slash_also_ok(webhook_testbot):
    assert requests.post(
        'http://localhost:{}/echo/'.format(WEBSERVER_PORT),
        JSONOBJECT
    ).text == repr(json.loads(JSONOBJECT))


def test_json_is_automatically_decoded(webhook_testbot):
    assert requests.post(
        'http://localhost:{}/webhook1'.format(WEBSERVER_PORT),
        JSONOBJECT
    ).text == repr(json.loads(JSONOBJECT))


def test_json_on_custom_url_is_automatically_decoded(webhook_testbot):
    assert requests.post(
        'http://localhost:{}/custom_webhook'.format(WEBSERVER_PORT),
        JSONOBJECT
    ).text == repr(json.loads(JSONOBJECT))


def test_post_form_on_webhook_without_form_param_is_automatically_decoded(webhook_testbot):
    assert requests.post(
        'http://localhost:{}/webhook1'.format(WEBSERVER_PORT),
        data=JSONOBJECT
    ).text == repr(json.loads(JSONOBJECT))


def test_post_form_on_webhook_with_custom_url_and_without_form_param_is_automatically_decoded(webhook_testbot):
    assert requests.post(
        'http://localhost:{}/custom_webhook'.format(WEBSERVER_PORT),
        data=JSONOBJECT
    ).text == repr(json.loads(JSONOBJECT))


def test_webhooks_with_form_parameter_decode_json_automatically(webhook_testbot):
    form = {'form': JSONOBJECT}
    assert requests.post(
        'http://localhost:{}/form'.format(WEBSERVER_PORT),
        data=form
    ).text == repr(json.loads(JSONOBJECT))


def test_webhooks_with_form_parameter_on_custom_url_decode_json_automatically(webhook_testbot):
    form = {'form': JSONOBJECT}
    assert requests.post(
        'http://localhost:{}/custom_form'.format(WEBSERVER_PORT),
        data=form
    ).text, repr(json.loads(JSONOBJECT))


def test_webhooks_with_raw_request(webhook_testbot):
    form = {'form': JSONOBJECT}
    assert 'LocalProxy' in requests.post('http://localhost:{}/raw'.format(WEBSERVER_PORT), data=form).text


def test_webhooks_with_naked_decorator_raw_request(webhook_testbot):
    form = {'form': JSONOBJECT}
    assert 'LocalProxy' in requests.post('http://localhost:{}/raw2'.format(WEBSERVER_PORT), data=form).text


def test_generate_certificate_creates_usable_cert(webhook_testbot):
    d = webhook_testbot.bot.bot_config.BOT_DATA_DIR
    key_path = os.sep.join((d, "webserver_key.pem"))
    cert_path = os.sep.join((d, "webserver_certificate.pem"))

    assert "Generating" in webhook_testbot.exec_command("!generate_certificate", timeout=1)

    # Generating a certificate could be slow on weak hardware, so keep a safe
    # timeout on the first pop_message()
    assert "successfully generated" in webhook_testbot.pop_message(timeout=60)
    assert "is recommended" in webhook_testbot.pop_message(timeout=1)
    assert key_path in webhook_testbot.pop_message(timeout=1)

    webserver_config = {
        'HOST': 'localhost',
        'PORT': WEBSERVER_PORT,
        'SSL': {
            'certificate': cert_path,
            'key': key_path,
            'host': 'localhost',
            'port': WEBSERVER_SSL_PORT,
            'enabled': True,
        }
    }
    webhook_testbot.push_message('!plugin config Webserver {!r}'.format(webserver_config))
    assert 'Plugin configuration done.' in webhook_testbot.pop_message(timeout=2)

    wait_for_server(WEBSERVER_SSL_PORT)

    assert requests.post(
        'https://localhost:{}/webhook1'.format(WEBSERVER_SSL_PORT),
        JSONOBJECT,
        verify=False
    ).text == repr(json.loads(JSONOBJECT))


def test_custom_headers_and_status_codes(webhook_testbot):
    assert requests.post(
        'http://localhost:{}/webhook6'.format(WEBSERVER_PORT)
    ).headers['X-Powered-By'] == 'Errbot'

    assert requests.post(
        'http://localhost:{}/webhook7'.format(WEBSERVER_PORT)
    ).status_code == 403


def test_lambda_webhook(webhook_testbot):
    assert requests.post(
        'http://localhost:{}/lambda'.format(WEBSERVER_PORT)
    ).status_code == 200



#!/usr/bin/env python

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import argparse
import locale
import logging
import os
import sys
from os import path, sep, getcwd, access, W_OK
from pathlib import Path
from platform import system
import ast

from errbot.bootstrap import CORE_BACKENDS
from errbot.logs import root_logger
from errbot.plugin_wizard import new_plugin_wizard
from errbot.utils import collect_roots
from errbot.version import VERSION

log = logging.getLogger(__name__)


# noinspection PyUnusedLocal
def debug(sig, frame):
    """Interrupt running process, and provide a python prompt for
    interactive debugging."""
    d = {'_frame': frame}  # Allow access to frame object.
    d.update(frame.f_globals)  # Unless shadowed by global
    d.update(frame.f_locals)

    i = code.InteractiveConsole(d)
    message = 'Signal received : entering python shell.\nTraceback:\n'
    message += ''.join(traceback.format_stack(frame))
    i.interact(message)


ON_WINDOWS = system() == 'Windows'

if not ON_WINDOWS:
    from daemonize import Daemonize
    import code
    import traceback
    import signal

    signal.signal(signal.SIGUSR1, debug)  # Register handler for debugging


def get_config(config_path):
    config_fullpath = config_path
    if not path.exists(config_fullpath):
        log.error(f'I cannot find the config file {config_path}.')
        log.error('You can change this path with the -c parameter see --help')
        log.info(f'You can use the template {os.path.realpath(os.path.join(__file__, os.pardir, "config-template.py"))}'
                 f' as a base and copy it to {config_path}.')
        log.info('You can then customize it.')
        exit(-1)

    try:
        config = __import__(path.splitext(path.basename(config_fullpath))[0])
        log.info('Config check passed...')
        return config
    except Exception:
        log.exception(f'I could not import your config from {config_fullpath}, please check the error below...')
        exit(-1)


def _read_dict():
    import collections
    new_dict = ast.literal_eval(sys.stdin.read())
    if not isinstance(new_dict, collections.Mapping):
        raise ValueError(f'A dictionary written in python is needed from stdin. '
                         f'Type={type(new_dict)}, Value = {repr(new_dict)}.')
    return new_dict


def main():

    execution_dir = getcwd()

    # By default insert the execution path (useful to be able to execute Errbot from
    # the source tree directly without installing it.
    sys.path.insert(0, execution_dir)

    parser = argparse.ArgumentParser(description='The main entry point of the errbot.')
    parser.add_argument('-c', '--config', default=None,
                        help='Full path to your config.py (default: config.py in current working directory).')

    mode_selection = parser.add_mutually_exclusive_group()
    mode_selection.add_argument('-v', '--version', action='version', version=f'Errbot version {VERSION}')
    mode_selection.add_argument('-r', '--restore', nargs='?', default=None, const='default',
                                help='restore a bot from backup.py (default: backup.py from the bot data directory)')
    mode_selection.add_argument('-l', '--list', action='store_true', help='list all available backends')
    mode_selection.add_argument('--new-plugin', nargs='?', default=None, const='current_dir',
                                help='create a new plugin in the specified directory')
    mode_selection.add_argument('-i', '--init',
                                nargs='?',
                                default=None,
                                const='.',
                                help='Initialize a simple bot minimal configuration in the optionally '
                                     'given directory (otherwise it will be the working directory). '
                                     'This will create a data subdirectory for the bot data dir and a plugins directory'
                                     ' for your plugin development with an example in it to get you started.')
    # storage manipulation
    mode_selection.add_argument('--storage-set', nargs=1, help='DANGER: Delete the given storage namespace '
                                                               'and set the python dictionary expression '
                                                               'passed on stdin.')
    mode_selection.add_argument('--storage-merge', nargs=1, help='DANGER: Merge in the python dictionary expression '
                                                                 'passed on stdin into the given storage namespace.')
    mode_selection.add_argument('--storage-get', nargs=1, help='Dump the given storage namespace in a '
                                                               'format compatible for --storage-set and '
                                                               '--storage-merge.')

    mode_selection.add_argument('-T', '--text', dest="backend", action='store_const', const="Text",
                                help='force local text backend')
    mode_selection.add_argument('-G', '--graphic', dest="backend", action='store_const', const="Graphic",
                                help='force local graphical backend')

    if not ON_WINDOWS:
        option_group = parser.add_argument_group('optional daemonization arguments')
        option_group.add_argument('-d', '--daemon', action='store_true', help='Detach the process from the console')
        option_group.add_argument('-p', '--pidfile', default=None,
                                  help='Specify the pid file for the daemon (default: current bot data directory)')

    args = vars(parser.parse_args())  # create a dictionary of args

    if args['init']:
        try:
            import jinja2
            import shutil
            import pathlib
            base_dir = pathlib.Path.cwd() if args['init'] == '.' else Path(args['init'])

            if not base_dir.exists():
                print(f'Target directory {base_dir} must exist. Please create it.')

            data_dir = base_dir / 'data'
            extra_plugin_dir = base_dir / 'plugins'
            example_plugin_dir = base_dir / extra_plugin_dir / 'err-example'
            log_path = base_dir / 'errbot.log'

            templates_dir = Path(os.path.dirname(__file__)) / 'templates' / 'initdir'
            env = jinja2.Environment(loader=jinja2.FileSystemLoader(str(templates_dir)), autoescape=True)
            config_template = env.get_template('config.py.tmpl')

            data_dir.mkdir(exist_ok=True)
            extra_plugin_dir.mkdir(exist_ok=True)
            example_plugin_dir.mkdir(exist_ok=True)

            with open(base_dir / 'config.py', 'w') as f:
                f.write(config_template.render(data_dir=str(data_dir),
                                               extra_plugin_dir=str(extra_plugin_dir),
                                               log_path=str(log_path)))

            shutil.copyfile(templates_dir / 'example.plug', example_plugin_dir / 'example.plug')
            shutil.copyfile(templates_dir / 'example.py', example_plugin_dir / 'example.py')
            print('Your Errbot directory has been correctly initialized!')
            if base_dir == pathlib.Path.cwd():
                print('Just do "errbot" and it should start in text/development mode.')
            else:
                print(f'Just do "cd {args["init"]}" then "errbot" and it should start in text/development mode.')
            sys.exit(0)
        except Exception as e:
            print(f'The initialization of your errbot directory failed: {e}.')
            sys.exit(1)

    # This must come BEFORE the config is loaded below, to avoid printing
    # logs as a side effect of config loading.
    if args['new_plugin']:
        directory = os.getcwd() if args['new_plugin'] == "current_dir" else args['new_plugin']
        for handler in logging.getLogger().handlers:
            root_logger.removeHandler(handler)
        try:
            new_plugin_wizard(directory)
        except KeyboardInterrupt:
            sys.exit(1)
        except Exception as e:
            sys.stderr.write(str(e) + "\n")
            sys.exit(1)
        finally:
            sys.exit(0)

    config_path = args['config']
    # setup the environment to be able to import the config.py
    if config_path:
        # appends the current config in order to find config.py
        sys.path.insert(0, path.dirname(path.abspath(config_path)))
    else:
        config_path = execution_dir + sep + 'config.py'

    config = get_config(config_path)  # will exit if load fails

    # Extra backend is expected to be a list type, convert string to list.
    extra_backend = getattr(config, 'BOT_EXTRA_BACKEND_DIR', [])
    if isinstance(extra_backend, str):
        extra_backend = [extra_backend]

    if args['list']:
        from errbot.backend_plugin_manager import enumerate_backend_plugins
        print('Available backends:')
        roots = [CORE_BACKENDS] + extra_backend
        for backend in enumerate_backend_plugins(collect_roots(roots)):
            print(f'\t\t{backend.name}')
        sys.exit(0)

    def storage_action(namespace, fn):
        # Used to defer imports until it is really necessary during the loading time.
        from errbot.bootstrap import get_storage_plugin
        from errbot.storage import StoreMixin
        try:
            with StoreMixin() as sdm:
                sdm.open_storage(get_storage_plugin(config), namespace)
                fn(sdm)
            return 0
        except Exception as e:
            print(str(e), file=sys.stderr)
            return -3

    if args['storage_get']:
        def p(sdm):
            print(repr(dict(sdm)))
        err_value = storage_action(args['storage_get'][0], p)
        sys.exit(err_value)

    if args['storage_set']:
        def replace(sdm):
            new_dict = _read_dict()  # fail early and don't erase the storage if the input is invalid.
            sdm.clear()
            sdm.update(new_dict)
        err_value = storage_action(args['storage_set'][0], replace)
        sys.exit(err_value)

    if args['storage_merge']:
        def merge(sdm):
            new_dict = _read_dict()
            if list(new_dict.keys()) == ['config']:
                with sdm.mutable('configs') as conf:
                    conf.update(new_dict['configs'])
            else:
                sdm.update(new_dict)
        err_value = storage_action(args['storage_merge'][0], merge)
        sys.exit(err_value)

    if args['restore']:
        backend = 'Null'  # we don't want any backend when we restore
    elif args['backend'] is None:
        if not hasattr(config, 'BACKEND'):
            log.fatal("The BACKEND configuration option is missing in config.py")
            sys.exit(1)
        backend = config.BACKEND
    else:
        backend = args['backend']

    log.info(f'Selected backend {backend}.')

    # Check if at least we can start to log something before trying to start
    # the bot (esp. daemonize it).

    log.info(f'Checking for {config.BOT_DATA_DIR}...')
    if not path.exists(config.BOT_DATA_DIR):
        raise Exception(f'The data directory "{config.BOT_DATA_DIR}" for the bot does not exist.')
    if not access(config.BOT_DATA_DIR, W_OK):
        raise Exception(f'The data directory "{config.BOT_DATA_DIR}" should be writable for the bot.')

    if (not ON_WINDOWS) and args['daemon']:
        if args['backend'] == 'Text':
            raise Exception('You cannot run in text and daemon mode at the same time')
        if args['restore']:
            raise Exception('You cannot restore a backup in daemon mode.')
        if args['pidfile']:
            pid = args['pidfile']
        else:
            pid = config.BOT_DATA_DIR + sep + 'err.pid'

        # noinspection PyBroadException
        try:
            def action():
                from errbot.bootstrap import bootstrap
                bootstrap(backend, root_logger, config)

            daemon = Daemonize(app="err", pid=pid, action=action, chdir=os.getcwd())
            log.info("Daemonizing")
            daemon.start()
        except Exception:
            log.exception('Failed to daemonize the process')
        exit(0)
    from errbot.bootstrap import bootstrap
    restore = args['restore']
    if restore == 'default':  # restore with no argument, get the default location
        restore = path.join(config.BOT_DATA_DIR, 'backup.py')

    bootstrap(backend, root_logger, config, restore)
    log.info('Process exiting')


if __name__ == "__main__":
    main()




import os
import io
from itertools import starmap, repeat
from threading import Thread
from .backends.base import STREAM_WAITING_TO_START, STREAM_TRANSFER_IN_PROGRESS
import logging

CHUNK_SIZE = 4096

log = logging.getLogger(__name__)


def repeatfunc(func, times=None, *args):  # from the itertools receipes
    """Repeat calls to func with specified arguments.

    Example:  repeatfunc(random.random)

    :param args: params to the function to call.
    :param times: number of times to repeat.
    :param func:  the function to repeatedly call.
    """
    if times is None:
        return starmap(func, repeat(args))
    return starmap(func, repeat(args, times))


class Tee(object):
    """ Tee implements a multi reader / single writer """
    def __init__(self, incoming_stream, clients):
        """ clients is a list of objects implementing callback_stream """
        self.incoming_stream = incoming_stream
        self.clients = clients

    def start(self):
        """ starts the transfer asynchronously """
        t = Thread(target=self.run)
        t.start()
        return t

    def run(self):
        """ streams to all the clients synchronously """
        nb_clients = len(self.clients)
        pipes = [(io.open(r, 'rb'), io.open(w, 'wb')) for r, w in repeatfunc(os.pipe, nb_clients)]
        streams = [self.incoming_stream.clone(pipe[0]) for pipe in pipes]

        def streamer(index):
            try:
                self.clients[index].callback_stream(streams[index])
                if streams[index].status == STREAM_WAITING_TO_START:
                    streams[index].reject()
                    plugin = self.clients[index].name
                    logging.warning('%s did not accept nor reject the incoming file transfer', plugin)
                    logging.warning('I reject it as a fallback.')
            except Exception as _:
                # internal error, mark the error.
                streams[index].error()
            else:
                if streams[index].status == STREAM_TRANSFER_IN_PROGRESS:
                    # if the plugin didn't do it by itself, mark the transfer as a success.
                    streams[index].success()
            # stop the stream if the callback_stream returns
            read, write = pipes[index]
            pipes[index] = (None, None)  # signal the main thread to stop streaming
            read.close()
            write.close()

        threads = [Thread(target=streamer, args=(i,)) for i in range(nb_clients)]

        for thread in threads:
            thread.start()

        while True:
            if self.incoming_stream.closed:
                break
            chunk = self.incoming_stream.read(CHUNK_SIZE)
            log.debug("dispatch %d bytes", len(chunk))
            if not chunk:
                break
            for (_, w) in pipes:
                if w:
                    w.write(chunk)
        log.debug("EOF detected")
        for (r, w) in pipes:
            if w:
                w.close()  # close should flush too
        # we want to be sure that if we join on the main thread,
        # everything is either fully transfered or errored
        for thread in threads:
            thread.join()



# coding=utf-8
from os import path

import pytest

extra_plugin_dir = path.join(path.dirname(path.realpath(__file__)), 'matchall_plugin')


def test_botmatch_correct(testbot):
    assert 'Works!' in testbot.exec_command('hi hi hi')


def test_botmatch(testbot):
    assert 'Works!' in testbot.exec_command('123123')



import sys
from tempfile import mkdtemp

import pytest
import logging
import os

from errbot.backends.text import TextBackend
from errbot.bootstrap import bot_config_defaults


@pytest.fixture
def text_backend():
    tempdir = mkdtemp()

    # reset the config every time
    sys.modules.pop('errbot.config-template', None)
    __import__('errbot.config-template')
    config = sys.modules['errbot.config-template']
    bot_config_defaults(config)
    config.BOT_DATA_DIR = tempdir
    config.BOT_LOG_FILE = os.path.join(tempdir, 'log.txt')
    config.BOT_EXTRA_PLUGIN_DIR = []
    config.BOT_LOG_LEVEL = logging.DEBUG
    config.BOT_PREFIX = '!'
    config.BOT_IDENTITY['username'] = '@testme'
    config.BOT_ADMINS = ['@test_admin']

    return TextBackend(config)


def test_change_presence(text_backend, caplog):
    with caplog.at_level(logging.DEBUG):
        text_backend.change_presence('online', "I'm online")
    assert len(caplog.messages) == 1
    a_message = caplog.messages[0]
    assert a_message.startswith('*** Changed presence')



# vim: ts=4:sw=4
import logging
from errbot import rendering

log = logging.getLogger(__name__)


def test_ansi():
    mdc = rendering.ansi()
    assert mdc.convert("*woot*") == "\x1b[4mwoot\x1b[24m\x1b[0m"


def test_text():
    mdc = rendering.text()
    assert mdc.convert("*woot*") == "woot"
    assert mdc.convert("# woot") == "WOOT"


def test_mde2md():
    mdc = rendering.md()
    assert mdc.convert("woot") == "woot"
    assert mdc.convert("woot{:stuff} really{:otherstuff}") == "woot really"


def test_escaping():
    mdc = rendering.text()
    original = '#not a title\n*not italic*\n`not code`\ntoto{not annotation}'
    escaped = rendering.md_escape(original)
    assert original == mdc.convert(escaped)



import argparse
from functools import wraps
import logging
import re
import shlex
import inspect
import sys
from typing import Callable, Any, Tuple

from .core_plugins.wsview import WebView
from .backends.base import Message, ONLINE, OFFLINE, AWAY, DND  # noqa
from .botplugin import BotPlugin, SeparatorArgParser, ShlexArgParser, CommandError, Command, ValidationException  # noqa
from .flow import FlowRoot, BotFlow, Flow, FLOW_END
from .core_plugins.wsview import route
from . import core

__all__ = ['BotPlugin', 'CommandError', 'Command', 'webhook', 'webroute', 'cmdfilter',
           'botcmd', 're_botcmd', 'arg_botcmd', 'botflow', 'botmatch', 'BotFlow', 'FlowRoot', 'Flow', 'FLOW_END',
           ]

log = logging.getLogger(__name__)

webroute = route  # this allows plugins to expose dynamic webpages on Errbot embedded webserver

# Some clients automatically convert consecutive dashes into a fancy
# hyphen, which breaks long-form arguments. Undo this conversion to
# provide a better user experience.
# Same happens with quotations marks, which are required for parsing
# complex strings in arguments
# Map of characters to sanitized equivalents
ARG_BOTCMD_CHARACTER_REPLACEMENTS = {'—': '--', '“': '"', '”': '"', '’': '\'', '‘': '\''}


class ArgumentParseError(Exception):
    """Raised when ArgumentParser couldn't parse given arguments."""


class HelpRequested(Exception):
    """Signals that -h/--help was used and help should be displayed to the user."""


class ArgumentParser(argparse.ArgumentParser):
    """
    The python argparse.ArgumentParser, adapted for use within Err.
    """

    def error(self, message):
        raise ArgumentParseError(message)

    def print_help(self, file=None):
        # Implementation note: Only easy way to do this appears to be
        #   through raising an exception which we can catch later in
        #   a place where we have the ability to return a message to
        #   the user.
        raise HelpRequested()


def _tag_botcmd(func,
                hidden=None,
                name=None,
                split_args_with='',
                admin_only=False,
                historize=True,
                template=None,
                flow_only=False,
                _re=False,
                syntax=None,  # botcmd_only
                pattern=None,  # re_cmd only
                flags=0,  # re_cmd only
                matchall=False,  # re_cmd_only
                prefixed=True,  # re_cmd_only
                _arg=False,
                command_parser=None,  # arg_cmd only
                re_cmd_name_help=None):  # re_cmd_only
    """
    Mark a method as a bot command.
    """
    if not hasattr(func, '_err_command'):  # don't override generated functions
        func._err_command = True
        func._err_command_name = name or func.__name__
        func._err_command_split_args_with = split_args_with
        func._err_command_admin_only = admin_only
        func._err_command_historize = historize
        func._err_command_template = template
        func._err_command_syntax = syntax
        func._err_command_flow_only = flow_only
        func._err_command_hidden = hidden if hidden is not None else flow_only

        # re_cmd
        func._err_re_command = _re
        if _re:
            func._err_command_re_pattern = re.compile(pattern, flags=flags)
            func._err_command_matchall = matchall
            func._err_command_prefix_required = prefixed
            func._err_command_syntax = pattern
            func._err_command_re_name_help = re_cmd_name_help

        # arg_cmd
        func._err_arg_command = _arg
        if _arg:
            func._err_command_parser = command_parser
            # func._err_command_syntax is set at wrapping time.
    return func


def botcmd(*args,
           hidden: bool = None,
           name: str = None,
           split_args_with: str = '',
           admin_only: bool = False,
           historize: bool = True,
           template: str = None,
           flow_only: bool = False,
           syntax: str = None) -> Callable[[BotPlugin, Message, Any], Any]:
    """
    Decorator for bot command functions

    :param hidden: Prevents the command from being shown by the built-in help command when `True`.
    :param name: The name to give to the command. Defaults to name of the function itself.
    :param split_args_with: Automatically split arguments on the given separator.
        Behaviour of this argument is identical to :func:`str.split()`
    :param admin_only: Only allow the command to be executed by admins when `True`.
    :param historize: Store the command in the history list (`!history`). This is enabled
        by default.
    :param template: The markdown template to use.
    :param syntax: The argument syntax you expect for example: '[name] <mandatory>'.
    :param flow_only: Flag this command to be available only when it is part of a flow.
                       If True and hidden is None, it will switch hidden to True.

    This decorator should be applied to methods of :class:`~errbot.botplugin.BotPlugin`
    classes to turn them into commands that can be given to the bot. These methods are
    expected to have a signature like the following::

        @botcmd
        def some_command(self, msg, args):
            pass

    The given `msg` will be the full message object that was received, which includes data
    like sender, receiver, the plain-text and html body (if applicable), etc. `args` will
    be a string or list (depending on your value of `split_args_with`) of parameters that
    were given to the command by the user.
    """

    def decorator(func):
        return _tag_botcmd(func,
                           _re=False,
                           _arg=False,
                           hidden=hidden,
                           name=name or func.__name__,
                           split_args_with=split_args_with,
                           admin_only=admin_only,
                           historize=historize,
                           template=template,
                           syntax=syntax,
                           flow_only=flow_only)

    return decorator(args[0]) if args else decorator


def re_botcmd(*args,
              hidden: bool = None,
              name: str = None,
              admin_only: bool = False,
              historize: bool = True,
              template: str = None,
              pattern: str = None,
              flags: int = 0,
              matchall: bool = False,
              prefixed: bool = True,
              flow_only: bool = False,
              re_cmd_name_help: str = None) -> Callable[[BotPlugin, Message, Any], Any]:
    """
    Decorator for regex-based bot command functions

    :param pattern: The regular expression a message should match against in order to
        trigger the command.
    :param flags: The `flags` parameter which should be passed to :func:`re.compile()`. This
        allows the expression's behaviour to be modified, such as making it case-insensitive
        for example.
    :param matchall: By default, only the first match of the regular expression is returned
        (as a `re.MatchObject`). When *matchall* is `True`, all non-overlapping matches are
        returned (as a list of `re.MatchObject` items).
    :param prefixed: Requires user input to start with a bot prefix in order for the pattern
        to be applied when `True` (the default).
    :param hidden: Prevents the command from being shown by the built-in help command when `True`.
    :param name: The name to give to the command. Defaults to name of the function itself.
    :param admin_only: Only allow the command to be executed by admins when `True`.
    :param historize: Store the command in the history list (`!history`). This is enabled
        by default.
    :param template: The template to use when using markdown output
    :param flow_only: Flag this command to be available only when it is part of a flow.
                       If True and hidden is None, it will switch hidden to True.

    This decorator should be applied to methods of :class:`~errbot.botplugin.BotPlugin`
    classes to turn them into commands that can be given to the bot. These methods are
    expected to have a signature like the following::

        @re_botcmd(pattern=r'^some command$')
        def some_command(self, msg, match):
            pass

    The given `msg` will be the full message object that was received, which includes data
    like sender, receiver, the plain-text and html body (if applicable), etc. `match` will
    be a :class:`re.MatchObject` containing the result of applying the regular expression on the
    user's input.
    """

    def decorator(func):
        return _tag_botcmd(func,
                           _re=True,
                           _arg=False,
                           hidden=hidden,
                           name=name or func.__name__,
                           admin_only=admin_only,
                           historize=historize,
                           template=template,
                           pattern=pattern,
                           flags=flags,
                           matchall=matchall,
                           prefixed=prefixed,
                           flow_only=flow_only,
                           re_cmd_name_help=re_cmd_name_help)

    return decorator(args[0]) if args else decorator


def botmatch(*args, **kwargs):
    """
    Decorator for regex-based message match.

    :param *args: The regular expression a message should match against in order to
                   trigger the command.
    :param flags: The `flags` parameter which should be passed to :func:`re.compile()`. This
        allows the expression's behaviour to be modified, such as making it case-insensitive
        for example.
    :param matchall: By default, only the first match of the regular expression is returned
        (as a `re.MatchObject`). When *matchall* is `True`, all non-overlapping matches are
        returned (as a list of `re.MatchObject` items).
    :param hidden: Prevents the command from being shown by the built-in help command when `True`.
    :param name: The name to give to the command. Defaults to name of the function itself.
    :param admin_only: Only allow the command to be executed by admins when `True`.
    :param historize: Store the command in the history list (`!history`). This is enabled
        by default.
    :param template: The template to use when using Markdown output.
    :param flow_only: Flag this command to be available only when it is part of a flow.
                       If True and hidden is None, it will switch hidden to True.

    For example::

        @botmatch(r'^(?:Yes|No)$')
        def yes_or_no(self, msg, match):
            pass
    """

    def decorator(func, pattern):
        return _tag_botcmd(func,
                           _re=True,
                           _arg=False,
                           prefixed=False,
                           hidden=kwargs.get('hidden', None),
                           name=kwargs.get('name', func.__name__),
                           admin_only=kwargs.get('admin_only', False),
                           flow_only=kwargs.get('flow_only', False),
                           historize=kwargs.get('historize', True),
                           template=kwargs.get('template', None),
                           pattern=pattern,
                           flags=kwargs.get('flags', 0),
                           matchall=kwargs.get('matchall', False))

    if len(args) == 2:
        return decorator(*args)
    if len(args) == 1:
        return lambda f: decorator(f, args[0])
    raise ValueError("botmatch: You need to pass the pattern as parameter to the decorator.")


def arg_botcmd(*args,
               hidden: bool = None,
               name: str = None,
               admin_only: bool = False,
               historize: bool = True,
               template: str = None,
               flow_only: bool = False,
               unpack_args: bool = True,
               **kwargs) -> Callable[[BotPlugin, Message, Any], Any]:
    """
    Decorator for argparse-based bot command functions

    https://docs.python.org/3/library/argparse.html

    This decorator creates an argparse.ArgumentParser and uses it to parse the commands arguments.

    This decorator can be used multiple times to specify multiple arguments.

    Any valid argparse.add_argument() parameters can be passed into the decorator.
    Each time this decorator is used it adds a new argparse argument to the command.

    :param hidden: Prevents the command from being shown by the built-in help command when `True`.
    :param name: The name to give to the command. Defaults to name of the function itself.
    :param admin_only: Only allow the command to be executed by admins when `True`.
    :param historize: Store the command in the history list (`!history`). This is enabled
        by default.
    :param template: The template to use when using markdown output
    :param flow_only: Flag this command to be available only when it is part of a flow.
                       If True and hidden is None, it will switch hidden to True.
    :param unpack_args: Should the argparser arguments be "unpacked" and passed on the the bot
        command individually? If this is True (the default) you must define all arguments in the
        function separately. If this is False you must define a single argument `args` (or
        whichever name you prefer) to receive the result of `ArgumentParser.parse_args()`.

    This decorator should be applied to methods of :class:`~errbot.botplugin.BotPlugin`
    classes to turn them into commands that can be given to the bot. The methods will be called
    with the original msg and the argparse parsed arguments. These methods are
    expected to have a signature like the following (assuming `unpack_args=True`)::

        @arg_botcmd('value', type=str)
        @arg_botcmd('--repeat-count', dest='repeat', type=int, default=2)
        def repeat_the_value(self, msg, value=None, repeat=None):
            return value * repeat

    The given `msg` will be the full message object that was received, which includes data
    like sender, receiver, the plain-text and html body (if applicable), etc.
    `value` will hold the value passed in place of the `value` argument and
    `repeat` will hold the value passed in place of the `--repeat-count` argument.

    If you don't like this automatic *"unpacking"* of the arguments,
    you can use `unpack_args=False` like this::

        @arg_botcmd('value', type=str)
        @arg_botcmd('--repeat-count', dest='repeat', type=int, default=2, unpack_args=False)
        def repeat_the_value(self, msg, args):
            return arg.value * args.repeat

    .. note::
        The `unpack_args=False` only needs to be specified once, on the bottom `@args_botcmd`
        statement.
    """
    argparse_args = args
    if len(args) >= 1 and callable(args[0]):
        argparse_args = args[1:]

    def decorator(func):

        if not hasattr(func, '_err_command'):

            err_command_parser = ArgumentParser(
                prog=name or func.__name__,
                description=func.__doc__,
            )

            @wraps(func)
            def wrapper(self, msg, args):

                # Attempt to sanitize arguments of bad characters
                try:
                    sanitizer_re = re.compile('|'.join(re.escape(ii) for ii in ARG_BOTCMD_CHARACTER_REPLACEMENTS))
                    args = sanitizer_re.sub(lambda mm: ARG_BOTCMD_CHARACTER_REPLACEMENTS[mm.group()], args)
                    args = shlex.split(args)
                    parsed_args = err_command_parser.parse_args(args)
                except ArgumentParseError as e:
                    yield f"I couldn't parse the arguments; {e}"
                    yield err_command_parser.format_usage()
                    return
                except HelpRequested:
                    yield err_command_parser.format_help()
                    return
                except ValueError as ve:
                    yield f"I couldn't parse this command; {ve}"
                    yield err_command_parser.format_help()
                    return

                if unpack_args:
                    func_args = []
                    func_kwargs = vars(parsed_args)
                else:
                    func_args = [parsed_args]
                    func_kwargs = {}

                if inspect.isgeneratorfunction(func):
                    for reply in func(self, msg, *func_args, **func_kwargs):
                        yield reply
                else:
                    yield func(self, msg, *func_args, **func_kwargs)

            _tag_botcmd(wrapper,
                        _re=False,
                        _arg=True,
                        hidden=hidden,
                        name=name or wrapper.__name__,
                        admin_only=admin_only,
                        historize=historize,
                        template=template,
                        flow_only=flow_only,
                        command_parser=err_command_parser)
        else:
            # the function has already been wrapped
            # alias it so we can update it's arguments below
            wrapper = func

        wrapper._err_command_parser.add_argument(*argparse_args, **kwargs)
        wrapper.__doc__ = wrapper._err_command_parser.format_help()
        fmt = wrapper._err_command_parser.format_usage()
        wrapper._err_command_syntax = fmt[len('usage: ') + len(wrapper._err_command_parser.prog) + 1:-1]

        return wrapper

    return decorator(args[0]) if callable(args[0]) else decorator


def _tag_webhook(func, uri_rule, methods, form_param, raw):
    log.info(f"webhooks:  Flag to bind {uri_rule} to {getattr(func, '__name__', func)}")
    func._err_webhook_uri_rule = uri_rule
    func._err_webhook_methods = methods
    func._err_webhook_form_param = form_param
    func._err_webhook_raw = raw
    return func


def _uri_from_func(func):
    return r'/' + func.__name__


def webhook(*args,
            methods: Tuple[str] = ('POST', 'GET'),
            form_param: str = None,
            raw: bool = False) -> Callable[[BotPlugin, Any], str]:
    """
    Decorator for webhooks

    :param uri_rule:
        The URL to use for this webhook, as per Flask request routing syntax.
        For more information, see:

        * http://flask.pocoo.org/docs/1.0/quickstart/#routing
        * http://flask.pocoo.org/docs/1.0/api/#flask.Flask.route
    :param methods:
        A tuple of allowed HTTP methods. By default, only GET and POST
        are allowed.
    :param form_param:
        The key who's contents will be passed to your method's `payload` parameter.
        This is used for example when using the `application/x-www-form-urlencoded`
        mimetype.
    :param raw:
        When set to true, this overrides the request decoding (including form_param) and
        passes the raw http request to your method's `payload` parameter.
        The value of payload will be a Flask
        `Request <http://flask.pocoo.org/docs/1.0/api/#flask.Request>`_.

    This decorator should be applied to methods of :class:`~errbot.botplugin.BotPlugin`
    classes to turn them into webhooks which can be reached on Err's built-in webserver.
    The bundled *Webserver* plugin needs to be configured before these URL's become reachable.

    Methods with this decorator are expected to have a signature like the following::

        @webhook
        def a_webhook(self, payload):
            pass
    """

    if not args:  # default uri_rule but with kwargs.
        return lambda func: _tag_webhook(func,
                                         _uri_from_func(func),
                                         methods=methods,
                                         form_param=form_param,
                                         raw=raw)

    if isinstance(args[0], str):  # first param is uri_rule.
        return lambda func: _tag_webhook(func,
                                         args[0] if args[0] == '/'
                                         else args[0].rstrip('/'),  # trailing / is also be stripped on incoming.
                                         methods=methods,
                                         form_param=form_param,
                                         raw=raw)
    return _tag_webhook(args[0],  # naked decorator so the first parameter is a function.
                        _uri_from_func(args[0]),
                        methods=methods,
                        form_param=form_param,
                        raw=raw)


def cmdfilter(*args, **kwargs):
    """
    Decorator for command filters.

    This decorator should be applied to methods of :class:`~errbot.botplugin.BotPlugin`
    classes to turn them into command filters.

    These filters are executed just before the execution of a command and provide
    the means to add features such as custom security, logging, auditing, etc.

    These methods are expected to have a signature and tuple response like the following::

        @cmdfilter
        def some_filter(self, msg, cmd, args, dry_run):
            \"\"\"
            :param msg: The original chat message.
            :param cmd: The command name itself.
            :param args: Arguments passed to the command.
            :param dry_run: True when this is a dry-run.
               Dry-runs are performed by certain commands (such as !help)
               to check whether a user is allowed to perform that command
               if they were to issue it. If dry_run is True then the plugin
               shouldn't actually do anything beyond returning whether the
               command is authorized or not.
            \"\"\"
            # If wishing to block the incoming command:
            return None, None, None
            # Otherwise pass data through to the (potential) next filter:
            return msg, cmd, args

    Note that a cmdfilter plugin *could* modify `cmd` or `args` above
    and send that through in order to make it appear as if the user
    issued a different command.
    """

    def decorate(func):
        if not hasattr(func, '_err_command_filter'):  # don't override generated functions
            func._err_command_filter = True
        func.catch_unprocessed = kwargs.get('catch_unprocessed', False)
        return func

    if len(args):
        return decorate(args[0])
    return lambda func: decorate(func)


def botflow(*args, **kwargs):
    """
    Decorator for flow of commands.

    TODO(gbin): example / docs
    """

    def decorate(func):
        if not hasattr(func, '_err_flow'):  # don't override generated functions
            func._err_flow = True
        return func

    if len(args):
        return decorate(args[0])
    return lambda func: decorate(func)



#!/usr/bin/env python

import errno
import jinja2
import os
import re
import sys
from configparser import ConfigParser

from errbot.version import VERSION


def new_plugin_wizard(directory=None):
    """
    Start the wizard to create a new plugin in the current working directory.
    """
    if directory is None:
        print('This wizard will create a new plugin for you in the current directory.')
        directory = os.getcwd()
    else:
        print(f'This wizard will create a new plugin for you in "{directory}".')

    if os.path.exists(directory) and not os.path.isdir(directory):
        print(f'Error: The path "{directory}" exists but it isn\'t a directory')
        sys.exit(1)

    name = ask("What should the name of your new plugin be?", validation_regex=r'^[a-zA-Z][a-zA-Z0-9 _-]*$').strip()
    module_name = name.lower().replace(' ', '_')
    directory_name = name.lower().replace(' ', '-')
    class_name = ''.join([s.capitalize() for s in name.lower().split(' ')])

    description = ask('What may I use as a short (one-line) description of your plugin?')
    python_version = '3'
    errbot_min_version = ask(f'Which minimum version of errbot will your plugin work with? '
                             f'Leave blank to support any version or input CURRENT to select '
                             f'the current version {VERSION}.').strip()
    if errbot_min_version.upper() == 'CURRENT':
        errbot_min_version = VERSION
    errbot_max_version = ask(f'Which maximum version of errbot will your plugin work with? '
                             f'Leave blank to support any version or input CURRENT to select '
                             f'the current version {VERSION}.').strip()
    if errbot_max_version.upper() == "CURRENT":
        errbot_max_version = VERSION

    plug = ConfigParser()
    plug['Core'] = {'Name': name,
                    'Module': module_name,
                    }
    plug['Documentation'] = {
        'Description': description,
    }
    plug['Python'] = {
        'Version': python_version,
    }

    if errbot_max_version != '' or errbot_min_version != '':
        plug['Errbot'] = {}
        if errbot_min_version != '':
            plug['Errbot']['Min'] = errbot_min_version
        if errbot_max_version != '':
            plug['Errbot']['Max'] = errbot_max_version

    plugin_path = directory
    plugfile_path = os.path.join(plugin_path, module_name + '.plug')
    pyfile_path = os.path.join(plugin_path, module_name + '.py')

    try:
        os.makedirs(plugin_path, mode=0o700)
    except IOError as e:
        if e.errno != errno.EEXIST:
            raise

    if os.path.exists(plugfile_path) or os.path.exists(pyfile_path):
        path = os.path.join(directory, f'{module_name}.{{py,plug}}')
        ask(f'Warning: A plugin with this name was already found at {path}\n'
            f'If you continue, these will be overwritten.\n'
            f'Press Ctrl+C to abort now or type in "overwrite" to confirm overwriting of these files.',
            valid_responses=['overwrite'],
            )

    with open(plugfile_path, 'w') as f:
        plug.write(f)

    with open(pyfile_path, 'w') as f:
        f.write(render_plugin(locals()))

    print(f'Success! You\'ll find your new plugin at \'{plugfile_path}\'')
    print('(Don\'t forget to include a LICENSE file if you are going to publish your plugin).')


def ask(question, valid_responses=None, validation_regex=None):
    """
    Ask the user for some input. If valid_responses is supplied, the user
    must respond with something present in this list.
    """
    response = None
    print(question)
    while True:
        response = input('> ')
        if valid_responses is not None:
            assert isinstance(valid_responses, list)
            if response in valid_responses:
                break
            else:
                print(f'Bad input: Please answer one of: {", ".join(valid_responses)}')
        elif validation_regex is not None:
            m = re.search(validation_regex, response)
            if m is None:
                print(f'Bad input: Please respond with something matching this regex: {validation_regex}')
            else:
                break
        else:
            break
    return response


def render_plugin(values):
    """
    Render the Jinja template for the plugin with the given values.
    """
    env = jinja2.Environment(
        loader=jinja2.FileSystemLoader(os.path.join(os.path.dirname(__file__), 'templates')),
        auto_reload=False,
        keep_trailing_newline=True,
        autoescape=True
    )
    template = env.get_template('new_plugin.py.tmpl')
    return template.render(**values)


if __name__ == '__main__':
    try:
        new_plugin_wizard()
    except KeyboardInterrupt:
        sys.exit(1)



import logging
import sys

from pathlib import Path
from typing import Any, Type

from errbot.plugin_info import PluginInfo
from .utils import collect_roots

log = logging.getLogger(__name__)


class PluginNotFoundException(Exception):
    pass


def enumerate_backend_plugins(all_plugins_paths):
    plugin_places = [Path(root) for root in all_plugins_paths]
    for path in plugin_places:
        plugfiles = path.glob('**/*.plug')
        for plugfile in plugfiles:
            plugin_info = PluginInfo.load(plugfile)
            yield plugin_info


class BackendPluginManager:
    """
    This is a one shot plugin manager for Backends and Storage plugins.
    """
    def __init__(self, bot_config, base_module: str, plugin_name: str, base_class: Type,
                 base_search_dir, extra_search_dirs=()):
        self._config = bot_config
        self._base_module = base_module
        self._base_class = base_class

        self.plugin_info = None
        all_plugins_paths = collect_roots((base_search_dir, extra_search_dirs))

        for potential_plugin in enumerate_backend_plugins(all_plugins_paths):
            if potential_plugin.name == plugin_name:
                self.plugin_info = potential_plugin
                return
        raise PluginNotFoundException(f'Could not find the plugin named {plugin_name} in {all_plugins_paths}.')

    def load_plugin(self) -> Any:
        plugin_path = self.plugin_info.location.parent
        if plugin_path not in sys.path:
            # Cast pathlib.Path objects to string type for compatibility with sys.path
            sys.path.append(str(plugin_path))
        plugin_classes = self.plugin_info.load_plugin_classes(self._base_module, self._base_class)
        if len(plugin_classes) != 1:
            raise PluginNotFoundException(f'Found more that one plugin for {self._base_class}.')

        _, clazz = plugin_classes[0]
        return clazz(self._config)



from os import path

from errbot import BotPlugin, botcmd


def tail(f, window=20):
    return ''.join(f.readlines()[-window:])


class Utils(BotPlugin):

    # noinspection PyUnusedLocal
    @botcmd
    def echo(self, _, args):
        """ A simple echo command. Useful for encoding tests etc ...
        """
        return args

    @botcmd
    def whoami(self, msg, args):
        """ A simple command echoing the details of your identifier. Useful to debug identity problems.
        """
        if args:
            frm = self.build_identifier(str(args).strip('"'))
        else:
            frm = msg.frm
        resp = "| key      | value\n"
        resp += "| -------- | --------\n"
        resp += f"| person   | `{frm.person}`\n"
        resp += f"| nick     | `{frm.nick}`\n"
        resp += f"| fullname | `{frm.fullname}`\n"
        resp += f"| client   | `{frm.client}`\n\n"

        #  extra info if it is a MUC
        if hasattr(frm, 'room'):
            resp += f"\n`room` is {frm.room}\n"
        resp += f"\n\n- string representation is '{frm}'\n"
        resp += f"- class is '{frm.__class__.__name__}'\n"

        return resp

    # noinspection PyUnusedLocal
    @botcmd(historize=False)
    def history(self, msg, args):
        """display the command history"""
        answer = []
        user_cmd_history = self._bot.cmd_history[msg.frm.person]
        length = len(user_cmd_history)
        for i in range(0, length):
            c = user_cmd_history[i]
            answer.append(f'{length - i:2d}:{self._bot.prefix}{c[0]} {c[1]}')
        return '\n'.join(answer)

    # noinspection PyUnusedLocal
    @botcmd(admin_only=True)
    def log_tail(self, msg, args):
        """ Display a tail of the log of n lines or 40 by default
        use : !log tail 10
        """
        n = 40
        if args.isdigit():
            n = int(args)

        if self.bot_config.BOT_LOG_FILE:
            with open(self.bot_config.BOT_LOG_FILE) as f:
                return '```\n' + tail(f, n) + '\n```'
        return 'No log is configured, please define BOT_LOG_FILE in config.py'

    @botcmd
    def render_test(self, _, args):
        """ Tests / showcases the markdown rendering on your current backend
        """
        with open(path.join(path.dirname(path.realpath(__file__)), 'test.md')) as f:
            return f.read()



import sys
import unittest
import logging
import os
from tempfile import mkdtemp
from mock import MagicMock

from errbot.bootstrap import bot_config_defaults

log = logging.getLogger(__name__)

try:
    from errbot.backends import slack

    class TestSlackBackend(slack.SlackBackend):

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.test_msgs = []
            self.sc = MagicMock()

        def callback_message(self, msg):
            self.test_msgs.append(msg)

        def username_to_userid(self, username, *args, **kwargs):
            """Have to mock because we don't have a slack server."""
            return 'Utest'

        def channelname_to_channelid(self, channelname):
            return 'Ctest'

        def channelid_to_channelname(self, channelid):
            return 'meh'

        def get_im_channel(self, id_):
            return 'Cfoo'

        def find_user(self, user):
            m = MagicMock()
            m.name = user
            return m

except SystemExit:
    log.exception("Can't import backends.slack for testing")


@unittest.skipIf(not slack, "package slackclient not installed")
class SlackTests(unittest.TestCase):

    def setUp(self):
        # make up a config.
        tempdir = mkdtemp()
        # reset the config every time
        sys.modules.pop('errbot.config-template', None)
        __import__('errbot.config-template')
        config = sys.modules['errbot.config-template']
        bot_config_defaults(config)
        config.BOT_DATA_DIR = tempdir
        config.BOT_LOG_FILE = os.path.join(tempdir, 'log.txt')
        config.BOT_EXTRA_PLUGIN_DIR = []
        config.BOT_LOG_LEVEL = logging.DEBUG
        config.BOT_IDENTITY = {'username': 'err@localhost', 'token': '___'}
        config.BOT_ASYNC = False
        config.BOT_PREFIX = '!'
        config.CHATROOM_FN = 'blah'

        self.slack = TestSlackBackend(config)

    def testBotMessageWithAttachments(self):
        attachment = {
            'title': 'sometitle',
            'id': 1,
            'fallback': ' *Host:* host-01', 'color': 'daa038',
            'fields': [{'title': 'Metric', 'value': '1', 'short': True}],
            'title_link': 'https://xx.com'
        }
        bot_id = 'B04HMXXXX'
        bot_msg = {
            'channel': 'C0XXXXY6P',
            'icons': {'emoji': ':warning:', 'image_64': 'https://xx.com/26a0.png'},
            'ts': '1444416645.000641',
            'type': 'message',
            'text': '',
            'bot_id': bot_id,
            'username': 'riemann',
            'subtype': 'bot_message',
            'attachments': [attachment]
        }

        self.slack._dispatch_slack_message(bot_msg)
        msg = self.slack.test_msgs.pop()

        self.assertEqual(msg.extras['attachments'], [attachment])

    def testSlackEventObjectAddedToExtras(self):
        bot_id = 'B04HMXXXX'
        bot_msg = {
            'channel': 'C0XXXXY6P',
            'icons': {'emoji': ':warning:', 'image_64': 'https://xx.com/26a0.png'},
            'ts': '1444416645.000641',
            'type': 'message',
            'text': '',
            'bot_id': bot_id,
            'username': 'riemann',
            'subtype': 'bot_message',
        }

        self.slack._dispatch_slack_message(bot_msg)
        msg = self.slack.test_msgs.pop()

        self.assertEqual(msg.extras['slack_event'], bot_msg)

    def testPrepareMessageBody(self):
        test_body = """
        hey, this is some code:
            ```
            foobar
            ```
        """
        parts = self.slack.prepare_message_body(test_body, 10000)
        assert parts == [test_body]

        test_body = """this block is unclosed: ``` foobar """
        parts = self.slack.prepare_message_body(test_body, 10000)
        assert parts == [test_body + "\n```\n"]

        test_body = """``` foobar """
        parts = self.slack.prepare_message_body(test_body, 10000)
        assert parts == [test_body + "\n```\n"]

        test_body = """closed ``` foobar ``` not closed ```"""
        # ---------------------------------^ 21st char
        parts = self.slack.prepare_message_body(test_body, 21)
        assert len(parts) == 2
        assert parts[0].count('```') == 2
        assert parts[0].endswith('```')
        assert parts[1].count('```') == 2
        assert parts[1].endswith('```\n')

    def test_extract_identifiers(self):
        extract_from = self.slack.extract_identifiers_from_string

        self.assertEqual(
            extract_from("<@U12345>"),
            (None, "U12345", None, None)
        )

        self.assertEqual(
            extract_from("<@U12345|UName>"),
            ("UName", "U12345", None, None)
        )

        self.assertEqual(
            extract_from("<@B12345>"),
            (None, "B12345", None, None)
        )

        self.assertEqual(
            extract_from("<#C12345>"),
            (None, None, None, "C12345")
        )

        self.assertEqual(
            extract_from("<#G12345>"),
            (None, None, None, "G12345")
        )

        self.assertEqual(
            extract_from("<#D12345>"),
            (None, None, None, "D12345")
        )

        self.assertEqual(
            extract_from("@person"),
            ("person", None, None, None)
        )

        self.assertEqual(
            extract_from("#general/someuser"),
            ("someuser", None, "general", None)
        )

        self.assertEqual(
            extract_from("#general"),
            (None, None, "general", None)
        )

        with self.assertRaises(ValueError):
            extract_from("")

        with self.assertRaises(ValueError):
            extract_from("general")

        with self.assertRaises(ValueError):
            extract_from("<>")

        with self.assertRaises(ValueError):
            extract_from("<C12345>")

        with self.assertRaises(ValueError):
            extract_from("<@I12345>")

    def test_build_identifier(self):
        build_from = self.slack.build_identifier

        def check_person(person, expected_uid, expected_cid):
            return person.userid == expected_uid and person.channelid == expected_cid
        assert build_from("<#C12345>").name == 'meh'
        assert check_person(build_from("<@U12345>"), "U12345", "Cfoo")
        assert check_person(build_from("@user"), "Utest", "Cfoo")
        assert build_from("#channel").name == 'meh'  # the mock always return meh ;)

        self.assertEqual(
            build_from("#channel/user"),
            slack.SlackRoomOccupant(None, "Utest", "Cfoo", self.slack)
        )

    def test_uri_sanitization(self):
        sanitize = self.slack.sanitize_uris

        self.assertEqual(
            sanitize(
                "The email is <mailto:test@example.org|test@example.org>."),
            "The email is test@example.org."
        )

        self.assertEqual(
            sanitize(
                "Pretty URL Testing: <http://example.org|example.org> with "
                "more text"),
            "Pretty URL Testing: example.org with more text"
        )

        self.assertEqual(
            sanitize("URL <http://example.org>"),
            "URL http://example.org"
        )

        self.assertEqual(
            sanitize("Normal &lt;text&gt; that shouldn't be affected"),
            "Normal &lt;text&gt; that shouldn't be affected"
        )

        self.assertEqual(
            sanitize(
                "Multiple uris <mailto:test@example.org|test@example.org>, "
                "<mailto:other@example.org|other@example.org> and "
                "<http://www.example.org>, <https://example.com> and "
                "<http://subdomain.example.org|subdomain.example.org>."),
            "Multiple uris test@example.org, other@example.org and "
            "http://www.example.org, https://example.com and subdomain.example.org."
        )

    def test_slack_markdown_link_preprocessor(self):
        convert = self.slack.md.convert
        self.assertEqual(
            "This is <http://example.com/|a link>.",
            convert("This is [a link](http://example.com/).")
        )
        self.assertEqual(
            "This is <https://example.com/|a link> and <mailto:me@comp.org|an email address>.",
            convert("This is [a link](https://example.com/) and [an email address](mailto:me@comp.org).")
        )
        self.assertEqual(
            "This is <http://example.com/|a link> and a manual URL: https://example.com/.",
            convert("This is [a link](http://example.com/) and a manual URL: https://example.com/.")
        )
        self.assertEqual(
            "<http://example.com/|This is a link>",
            convert("[This is a link](http://example.com/)")
        )
        self.assertEqual(
            "This is http://example.com/image.png.",
            convert("This is ![an image](http://example.com/image.png).")
        )
        self.assertEqual(
            "This is [some text] then <http://example.com|a link>",
            convert("This is [some text] then [a link](http://example.com)")
        )

    def test_mention_processing(self):
        self.slack.sc.server.users.find = MagicMock(side_effect=self.slack.find_user)

        mentions = self.slack.process_mentions

        self.assertEqual(
            mentions(
                "<@U1><@U2><@U3>"),
            (
                "@U1@U2@U3",
                [self.slack.build_identifier('<@U1>'),
                 self.slack.build_identifier('<@U2>'),
                 self.slack.build_identifier('<@U3>')])
        )

        self.assertEqual(
            mentions(
                "Is <@U12345>: here?"),
            (
                "Is @U12345: here?", [self.slack.build_identifier('<@U12345>')])
        )

        self.assertEqual(
            mentions(
                "<@U12345> told me about @a and <@U56789> told me about @b"),
            (
                "@U12345 told me about @a and @U56789 told me about @b",
                [self.slack.build_identifier('<@U12345>'),
                 self.slack.build_identifier('<@U56789>')])
        )

        self.assertEqual(
            mentions(
                "!these!<@UABCDE>!mentions! will !still!<@UFGHIJ>!work!"),
            (
                "!these!@UABCDE!mentions! will !still!@UFGHIJ!work!",
                [self.slack.build_identifier('<@UABCDE>'),
                 self.slack.build_identifier('<@UFGHIJ>')])
        )



from __future__ import absolute_import
from errbot import BotPlugin, botcmd, re_botcmd, arg_botcmd


class Test(BotPlugin):
    """Just a test plugin to see if _err_botcmd_syntax is consistent on all types of botcmd
    """
    @botcmd     # no syntax
    def foo_nosyntax(self, msg, args):
        pass

    @botcmd(syntax='[optional] <mandatory>')
    def foo(self, msg, args):
        pass

    @re_botcmd(pattern=r".*")
    def re_foo(self, msg, match):
        pass

    @arg_botcmd('value', type=str)
    @arg_botcmd('--repeat-count', dest='repeat', type=int, default=2)
    def arg_foo(self, msg, value=None, repeat=None):
        return value * repeat



""" Logic related to plugin loading and lifecycle """
from copy import deepcopy
from importlib import machinery
import logging
import os
import subprocess
import sys
import traceback
from pathlib import Path

from typing import Tuple, Dict, Any, Type, Set, List, Optional, Callable

from errbot.flow import BotFlow, Flow
from errbot.repo_manager import check_dependencies
from errbot.storage.base import StoragePluginBase
from .botplugin import BotPlugin
from .plugin_info import PluginInfo
from .utils import version2tuple, collect_roots
from .templating import remove_plugin_templates_path, add_plugin_templates_path
from .version import VERSION
from .core_plugins.wsview import route
from .storage import StoreMixin

PluginInstanceCallback = Callable[[str, Type[BotPlugin]], BotPlugin]

log = logging.getLogger(__name__)

CORE_PLUGINS = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'core_plugins')


class PluginActivationException(Exception):
    pass


class IncompatiblePluginException(PluginActivationException):
    pass


class PluginConfigurationException(PluginActivationException):
    pass


def _ensure_sys_path_contains(paths):
    """ Ensure that os.path contains paths
       :param paths:
            a list of base paths to walk from
            elements can be a string or a list/tuple of strings
    """
    for entry in paths:
        if isinstance(entry, (list, tuple)):
            _ensure_sys_path_contains(entry)
        elif entry is not None and entry not in sys.path:
            sys.path.append(entry)


def populate_doc(plugin_object: BotPlugin, plugin_info: PluginInfo) -> None:
    plugin_class = type(plugin_object)
    plugin_class.__errdoc__ = plugin_class.__doc__ if plugin_class.__doc__ else plugin_info.doc


def install_packages(req_path: Path):
    """ Installs all the packages from the given requirements.txt

        Return an exc_info if it fails otherwise None.
    """
    def is_docker():
        if not os.path.exists('/proc/1/cgroup'):
            return false
        with open('/proc/1/cgroup') as d:
            return 'docker' in d.read()

    log.info('Installing packages from "%s".', req_path)
    # use sys.executable explicitly instead of just 'pip' because depending on how the bot is deployed
    # 'pip' might not be available on PATH: for example when installing errbot on a virtualenv and
    # starting it with systemclt pointing directly to the executable:
    # [Service]
    # ExecStart=/home/errbot/.env/bin/errbot
    pip_cmdline = [sys.executable, '-m', 'pip']
    # noinspection PyBroadException
    try:
        if hasattr(sys, 'real_prefix') or (hasattr(sys, 'base_prefix') and (sys.base_prefix != sys.prefix)):
            # this is a virtualenv, so we can use it directly
            subprocess.check_call(pip_cmdline + ['install', '--requirement', str(req_path)])
        elif is_docker():
            # this is a docker container, so we can use it directly
            subprocess.check_call(pip_cmdline + ['install', '--requirement', str(req_path)])
        else:
            # otherwise only install it as a user package
            subprocess.check_call(pip_cmdline + ['install', '--user', '--requirement', str(req_path)])
    except Exception:
        log.exception('Failed to execute pip install for %s.', req_path)
        return sys.exc_info()


def check_python_plug_section(plugin_info: PluginInfo) -> bool:
    """ Checks if we have the correct version to run this plugin.
    Returns true if the plugin is loadable """
    version = plugin_info.python_version

    # if the plugin doesn't restric anything, assume it is ok and try to load it.
    if not version:
        return True

    sys_version = sys.version_info[:3]
    if version < (3, 0, 0):
        log.error('Plugin %s is made for python 2 only and Errbot is not compatible with Python 2 anymore.',
                  plugin_info.name)
        log.error('Please contact the plugin developer or try to contribute to port the plugin.')
        return False

    if version >= sys_version:
        log.error('Plugin %s requires python >= %s and this Errbot instance runs %s.',
                  plugin_info.name, '.'.join(str(v) for v in version), '.'.join(str(v) for v in sys_version))
        log.error('Upgrade your python interpreter if you want to use this plugin.')
        return False

    return True


def check_errbot_version(plugin_info: PluginInfo):
    """ Checks if a plugin version between min_version and max_version is ok
    for this errbot.
    Raises IncompatiblePluginException if not.
    """
    name, min_version, max_version = plugin_info.name, plugin_info.errbot_minversion, plugin_info.errbot_maxversion
    current_version = version2tuple(VERSION)
    if min_version and min_version > current_version:
        raise IncompatiblePluginException(f'The plugin {name} asks for Errbot with a minimal version of '
                                          f'{min_version} while Errbot is version {VERSION}.')

    if max_version and max_version < current_version:
        raise IncompatiblePluginException(f'The plugin {name} asks for Errbot with a maximum version of {max_version} '
                                          f'while Errbot is version {VERSION}')


# Storage names
CONFIGS = 'configs'
BL_PLUGINS = 'bl_plugins'


class BotPluginManager(StoreMixin):

    def __init__(self,
                 storage_plugin: StoragePluginBase,
                 extra_plugin_dir: Optional[str],
                 autoinstall_deps: bool,
                 core_plugins: Tuple[str, ...],
                 plugin_instance_callback: PluginInstanceCallback,
                 plugins_callback_order: Tuple[Optional[str], ...]):
        """
        Creates a Plugin manager
        :param storage_plugin: the plugin used to store to config for this manager
        :param extra_plugin_dir: an extra directory to search for plugins
        :param autoinstall_deps: if True, will install also the plugin deps from requirements.txt
        :param core_plugins: the list of core plugin that will be started
        :param plugin_instance_callback: the callback to instantiate a plugin (to inject the dependency on the bot)
        :param plugins_callback_order: the order on which the plugins will be callbacked
        """
        super().__init__()
        self.autoinstall_deps: bool = autoinstall_deps
        self._extra_plugin_dir: str = extra_plugin_dir
        self._plugin_instance_callback: PluginInstanceCallback = plugin_instance_callback
        self.core_plugins: Tuple[str, ...] = core_plugins
        # Make sure there is a 'None' entry in the callback order, to include
        # any plugin not explicitly ordered.
        self.plugins_callback_order = plugins_callback_order
        if None not in self.plugins_callback_order:
            self.plugins_callback_order += (None,)
        self.plugin_infos: Dict[str, PluginInfo] = {}
        self.plugins: Dict[str, BotPlugin] = {}
        self.flow_infos: Dict[str, PluginInfo] = {}
        self.flows: Dict[str, Flow] = {}
        self.plugin_places = []
        self.open_storage(storage_plugin, 'core')
        if CONFIGS not in self:
            self[CONFIGS] = {}

    def get_plugin_obj_by_name(self, name: str) -> BotPlugin:
        return self.plugins.get(name, None)

    def reload_plugin_by_name(self, name):
        """
        Completely reload the given plugin, including reloading of the module's code
        :throws PluginActivationException: needs to be taken care of by the callers.
        """
        plugin = self.plugins[name]
        plugin_info = self.plugin_infos[name]

        if plugin.is_activated:
            self.deactivate_plugin(name)

        base_name = '.'.join(plugin.__module__.split('.')[:-1])
        classes = plugin_info.load_plugin_classes(base_name, BotPlugin)
        _, new_class = classes[0]
        plugin.__class__ = new_class

        self.activate_plugin(name)

    def _install_potential_package_dependencies(self, path: Path,
                                                feedback: Dict[Path, str]):
        req_path = path / 'requirements.txt'
        if req_path.exists():
            log.info('Checking package dependencies from %s.', req_path)
            if self.autoinstall_deps:
                exc_info = install_packages(req_path)
                if exc_info is not None:
                    typ, value, trace = exc_info
                    feedback[path] = f'{typ}: {value}\n{"".join(traceback.format_tb(trace))}'
            else:
                msg, _ = check_dependencies(req_path)
                if msg and path not in feedback:  # favor the first error.
                    feedback[path] = msg

    def _load_plugins_generic(self,
                              path: Path,
                              extension: str,
                              base_module_name,
                              baseclass: Type,
                              dest_dict: Dict[str, Any],
                              dest_info_dict: Dict[str, Any],
                              feedback: Dict[Path, str]):
        self._install_potential_package_dependencies(path, feedback)
        plugfiles = path.glob('**/*.' + extension)
        for plugfile in plugfiles:
            try:
                plugin_info = PluginInfo.load(plugfile)
                name = plugin_info.name
                if name in dest_info_dict:
                    log.warning('Plugin %s already loaded.', name)
                    continue

                # save the plugin_info for ref.
                dest_info_dict[name] = plugin_info

                # Skip the core plugins not listed in CORE_PLUGINS if CORE_PLUGINS is defined.
                if self.core_plugins and plugin_info.core and (plugin_info.name not in self.core_plugins):
                    log.debug("%s plugin will not be loaded because it's not listed in CORE_PLUGINS", name)
                    continue

                plugin_classes = plugin_info.load_plugin_classes(base_module_name, baseclass)
                if not plugin_classes:
                    feedback[path] = f'Did not find any plugin in {path}.'
                    continue
                if len(plugin_classes) > 1:
                    # TODO: This is something we can support as "subplugins" or something similar.
                    feedback[path] = 'Contains more than one plugin, only one will be loaded.'

                # instantiate the plugin object.
                _, clazz = plugin_classes[0]
                dest_dict[name] = self._plugin_instance_callback(name, clazz)

            except Exception:
                feedback[path] = traceback.format_exc()

    def _load_plugins(self) -> Dict[Path, str]:
        feedback = {}
        for path in self.plugin_places:
            self._load_plugins_generic(path, 'plug', 'errbot.plugins', BotPlugin,
                                       self.plugins, self.plugin_infos, feedback)
            self._load_plugins_generic(path, 'flow', 'errbot.flows', BotFlow,
                                       self.flows, self.flow_infos, feedback)
        return feedback

    def update_plugin_places(self, path_list) -> Dict[Path, str]:
        """
        This updates where this manager is trying to find plugins and try to load newly found ones.
        :param path_list: the path list where to search for plugins.
        :return: the feedback for any specific path in case of error.
        """
        repo_roots = (CORE_PLUGINS, self._extra_plugin_dir, path_list)

        all_roots = collect_roots(repo_roots)

        log.debug('New entries added to sys.path:')
        for entry in all_roots:
            if entry not in sys.path:
                log.debug(entry)
                sys.path.append(entry)
        # so plugins can relatively import their repos
        _ensure_sys_path_contains(repo_roots)
        self.plugin_places = [Path(root) for root in all_roots]
        return self._load_plugins()

    def get_all_active_plugins(self) -> List[BotPlugin]:
        """This returns the list of plugins in the callback ordered defined from the config."""

        all_plugins = []
        for name in self.plugins_callback_order:
            # None is a placeholder for any plugin not having a defined order
            if name is None:
                all_plugins += [
                    plugin for name, plugin in self.plugins.items()
                    if name not in self.plugins_callback_order and plugin.is_activated
                ]
            else:
                plugin = self.plugins[name]
                if plugin.is_activated:
                    all_plugins.append(plugin)
        return all_plugins

    def get_all_active_plugin_names(self):
        return [name for name, plugin in self.plugins.items() if plugin.is_activated]

    def get_all_plugin_names(self):
        return self.plugins.keys()

    def get_plugin_by_path(self, path):
        for name, pi in self.plugin_infos.items():
            if str(pi.location.parent) == path:
                return self.plugins[name]

    def deactivate_all_plugins(self):
        for name in self.get_all_active_plugin_names():
            self.deactivate_plugin(name)

    # plugin blacklisting management
    def get_blacklisted_plugin(self):
        return self.get(BL_PLUGINS, [])

    def is_plugin_blacklisted(self, name):
        return name in self.get_blacklisted_plugin()

    def blacklist_plugin(self, name):
        if self.is_plugin_blacklisted(name):
            logging.warning('Plugin %s is already blacklisted.', name)
            return f'Plugin {name} is already blacklisted.'
        self[BL_PLUGINS] = self.get_blacklisted_plugin() + [name]
        log.info('Plugin %s is now blacklisted.', name)
        return f'Plugin {name} is now blacklisted.'

    def unblacklist_plugin(self, name):
        if not self.is_plugin_blacklisted(name):
            logging.warning('Plugin %s is not blacklisted.', name)
            return f'Plugin {name} is not blacklisted.'
        plugin = self.get_blacklisted_plugin()
        plugin.remove(name)
        self[BL_PLUGINS] = plugin
        log.info('Plugin %s removed from blacklist.', name)
        return f'Plugin {name} removed from blacklist.'

    # configurations management
    def get_plugin_configuration(self, name):
        configs = self[CONFIGS]
        if name not in configs:
            return None
        return configs[name]

    def set_plugin_configuration(self, name, obj):
        # TODO: port to with statement
        configs = self[CONFIGS]
        configs[name] = obj
        self[CONFIGS] = configs

    def activate_non_started_plugins(self):
        """
        Activates all plugins that are not activated, respecting its dependencies.

        :return: Empty string if no problem occured or a string explaining what went wrong.
        """
        log.info('Activate bot plugins...')
        errors = ''
        for name, plugin in self.plugins.items():
            try:
                if self.is_plugin_blacklisted(name):
                    errors += f'Notice: {plugin.name} is blacklisted, ' \
                              f'use {self.bot.prefix}plugin unblacklist {name} to unblacklist it.\n'
                    continue
                if not plugin.is_activated:
                    log.info('Activate plugin: %s.', name)
                    self.activate_plugin(name)
            except Exception as e:
                log.exception('Error loading %s.', name)
                errors += f'Error: {name} failed to activate: {e}.\n'

        log.debug('Activate flow plugins ...')
        for name, flow in self.flows.items():
            try:
                if not flow.is_activated:
                    log.info('Activate flow: %s', name)
                    self.activate_flow(name)
            except Exception as e:
                log.exception(f'Error loading flow {name}.')
                errors += f'Error: flow {name} failed to start: {e}.\n'
        return errors

    def _activate_plugin(self, plugin: BotPlugin, plugin_info: PluginInfo):
        """
        Activate a specific plugin with no check.
        """
        if plugin.is_activated:
            raise Exception('Internal Error, invalid activated state.')

        name = plugin.name
        try:
            config = self.get_plugin_configuration(name)
            if plugin.get_configuration_template() is not None and config is not None:
                log.debug('Checking configuration for %s...', name)
                plugin.check_configuration(config)
                log.debug('Configuration for %s checked OK.', name)
            plugin.configure(config)  # even if it is None we pass it on
        except Exception as ex:
            log.exception('Something is wrong with the configuration of the plugin %s', name)
            plugin.config = None
            raise PluginConfigurationException(str(ex))

        try:
            add_plugin_templates_path(plugin_info)
            populate_doc(plugin, plugin_info)
            plugin.activate()
            route(plugin)
            plugin.callback_connect()
        except Exception:
            log.error('Plugin %s failed at activation stage, deactivating it...', name)
            self.deactivate_plugin(name)
            raise

    def activate_flow(self, name: str):
        if name not in self.flows:
            raise PluginActivationException(f'Could not find the flow named {name}.')

        flow = self.flows[name]
        if flow.is_activated:
            raise PluginActivationException(f'Flow {name} is already active.')
        flow.activate()

    def deactivate_flow(self, name: str):
        flow = self.flows[name]
        if not flow.is_activated:
            raise PluginActivationException(f'Flow {name} is already inactive.')
        flow.deactivate()

    def activate_plugin(self, name: str):
        """
        Activate a plugin with its dependencies.
        """
        try:
            if name not in self.plugins:
                raise PluginActivationException(f'Could not find the plugin named {name}.')

            plugin = self.plugins[name]
            if plugin.is_activated:
                raise PluginActivationException(f'Plugin {name} already activate.')

            plugin_info = self.plugin_infos[name]

            if not check_python_plug_section(plugin_info):
                return None

            check_errbot_version(plugin_info)

            dep_track = set()
            depends_on = self._activate_plugin_dependencies(name, dep_track)
            plugin.dependencies = depends_on
            self._activate_plugin(plugin, plugin_info)

        except PluginActivationException:
            raise
        except Exception as e:
            log.exception(f'Error loading {name}.')
            raise PluginActivationException(f'{name} failed to start : {e}.')

    def _activate_plugin_dependencies(self, name: str, dep_track: Set[str]) -> List[str]:

        plugin_info = self.plugin_infos[name]
        dep_track.add(name)

        depends_on = plugin_info.dependencies
        for dep_name in depends_on:
            if dep_name in dep_track:
                raise PluginActivationException(f'Circular dependency in the set of plugins ({", ".join(dep_track)})')
            if dep_name not in self.plugins:
                raise PluginActivationException(f'Unknown plugin dependency {dep_name}.')
            dep_plugin = self.plugins[dep_name]
            dep_plugin_info = self.plugin_infos[dep_name]
            if not dep_plugin.is_activated:
                log.debug('%s depends on %s and %s is not activated. Activating it ...', name, dep_name, dep_name)
                self._activate_plugin_dependencies(dep_name, dep_track)
                self._activate_plugin(dep_plugin, dep_plugin_info)
        return depends_on

    def deactivate_plugin(self, name: str):
        plugin = self.plugins[name]
        if not plugin.is_activated:
            log.warning('Plugin already deactivated, ignore.')
            return
        plugin_info = self.plugin_infos[name]
        plugin.deactivate()
        remove_plugin_templates_path(plugin_info)

    def remove_plugin(self, plugin: BotPlugin):
        """
        Deactivate and remove a plugin completely.
        :param plugin: the plugin to remove
        :return:
        """
        # First deactivate it if it was activated
        if plugin.is_activated:
            self.deactivate_plugin(plugin.name)

        del(self.plugins[plugin.name])
        del(self.plugin_infos[plugin.name])

    def remove_plugins_from_path(self, root):
        """
        Remove all the plugins that are in the filetree pointed by root.
        """
        old_plugin_infos = deepcopy(self.plugin_infos)
        for name, pi in old_plugin_infos.items():
            if str(pi.location).startswith(root):
                self.remove_plugin(self.plugins[name])

    def shutdown(self):
        log.info('Shutdown.')
        self.close_storage()
        log.info('Bye.')

    def __hash__(self):
        # Ensures this class (and subclasses) are hashable.
        # Presumably the use of mixins causes __hash__ to be
        # None otherwise.
        return int(id(self))



from os import path, makedirs
import importlib
import logging
import sys

from errbot.core import ErrBot
from errbot.plugin_manager import BotPluginManager
from errbot.repo_manager import BotRepoManager
from errbot.backend_plugin_manager import BackendPluginManager
from errbot.storage.base import StoragePluginBase
from errbot.utils import PLUGINS_SUBDIR
from errbot.logs import format_logs

log = logging.getLogger(__name__)

HERE = path.dirname(path.abspath(__file__))
CORE_BACKENDS = path.join(HERE, 'backends')
CORE_STORAGE = path.join(HERE, 'storage')

PLUGIN_DEFAULT_INDEX = 'https://errbot.io/repos.json'


def bot_config_defaults(config):
    if not hasattr(config, 'ACCESS_CONTROLS_DEFAULT'):
        config.ACCESS_CONTROLS_DEFAULT = {}
    if not hasattr(config, 'ACCESS_CONTROLS'):
        config.ACCESS_CONTROLS = {}
    if not hasattr(config, 'HIDE_RESTRICTED_COMMANDS'):
        config.HIDE_RESTRICTED_COMMANDS = False
    if not hasattr(config, 'HIDE_RESTRICTED_ACCESS'):
        config.HIDE_RESTRICTED_ACCESS = False
    if not hasattr(config, 'BOT_PREFIX_OPTIONAL_ON_CHAT'):
        config.BOT_PREFIX_OPTIONAL_ON_CHAT = False
    if not hasattr(config, 'BOT_PREFIX'):
        config.BOT_PREFIX = '!'
    if not hasattr(config, 'BOT_ALT_PREFIXES'):
        config.BOT_ALT_PREFIXES = ()
    if not hasattr(config, 'BOT_ALT_PREFIX_SEPARATORS'):
        config.BOT_ALT_PREFIX_SEPARATORS = ()
    if not hasattr(config, 'BOT_ALT_PREFIX_CASEINSENSITIVE'):
        config.BOT_ALT_PREFIX_CASEINSENSITIVE = False
    if not hasattr(config, 'DIVERT_TO_PRIVATE'):
        config.DIVERT_TO_PRIVATE = ()
    if not hasattr(config, 'DIVERT_TO_THREAD'):
        config.DIVERT_TO_THREAD = ()
    if not hasattr(config, 'MESSAGE_SIZE_LIMIT'):
        config.MESSAGE_SIZE_LIMIT = 10000  # Corresponds with what HipChat accepts
    if not hasattr(config, 'GROUPCHAT_NICK_PREFIXED'):
        config.GROUPCHAT_NICK_PREFIXED = False
    if not hasattr(config, 'AUTOINSTALL_DEPS'):
        config.AUTOINSTALL_DEPS = True
    if not hasattr(config, 'SUPPRESS_CMD_NOT_FOUND'):
        config.SUPPRESS_CMD_NOT_FOUND = False
    if not hasattr(config, 'BOT_ASYNC'):
        config.BOT_ASYNC = True
    if not hasattr(config, 'BOT_ASYNC_POOLSIZE'):
        config.BOT_ASYNC_POOLSIZE = 10
    if not hasattr(config, 'CHATROOM_PRESENCE'):
        config.CHATROOM_PRESENCE = ()
    if not hasattr(config, 'CHATROOM_RELAY'):
        config.CHATROOM_RELAY = ()
    if not hasattr(config, 'REVERSE_CHATROOM_RELAY'):
        config.REVERSE_CHATROOM_RELAY = ()
    if not hasattr(config, 'CHATROOM_FN'):
        config.CHATROOM_FN = 'Errbot'
    if not hasattr(config, 'TEXT_DEMO_MODE'):
        config.TEXT_DEMO_MODE = True
    if not hasattr(config, 'BOT_ADMINS'):
        raise ValueError('BOT_ADMINS missing from config.py.')
    if not hasattr(config, 'TEXT_COLOR_THEME'):
        config.TEXT_COLOR_THEME = 'light'
    if not hasattr(config, 'BOT_ADMINS_NOTIFICATIONS'):
        config.BOT_ADMINS_NOTIFICATIONS = config.BOT_ADMINS


def setup_bot(backend_name: str, logger, config, restore=None) -> ErrBot:
    # from here the environment is supposed to be set (daemon / non daemon,
    # config.py in the python path )

    bot_config_defaults(config)

    if hasattr(config, 'BOT_LOG_FORMATTER'):
        format_logs(formatter=config.BOT_LOG_FORMATTER)
    else:
        format_logs(theme_color=config.TEXT_COLOR_THEME)

        if config.BOT_LOG_FILE:
            hdlr = logging.FileHandler(config.BOT_LOG_FILE)
            hdlr.setFormatter(logging.Formatter("%(asctime)s %(levelname)-8s %(name)-25s %(message)s"))
            logger.addHandler(hdlr)

    if hasattr(config, 'BOT_LOG_SENTRY') and config.BOT_LOG_SENTRY:
        sentry_integrations = []

        try:
            import sentry_sdk
            from sentry_sdk.integrations.logging import LoggingIntegration

        except ImportError:
            log.exception(
                "You have BOT_LOG_SENTRY enabled, but I couldn't import modules "
                "needed for Sentry integration. Did you install sentry-sdk? "
                "(See https://docs.sentry.io/platforms/python for installation instructions)"
            )
            exit(-1)

        sentry_logging = LoggingIntegration(
            level=config.SENTRY_LOGLEVEL,
            event_level=config.SENTRY_EVENTLEVEL
        )

        sentry_integrations.append(sentry_logging)

        if hasattr(config, 'BOT_LOG_SENTRY_FLASK') and config.BOT_LOG_SENTRY_FLASK:
            try:
                from sentry_sdk.integrations.flask import FlaskIntegration
            except ImportError:
                log.exception(
                    "You have BOT_LOG_SENTRY enabled, but I couldn't import modules "
                    "needed for Sentry integration. Did you install sentry-sdk[flask]? "
                    "(See https://docs.sentry.io/platforms/python/flask for installation instructions)"
                )
                exit(-1)

            sentry_integrations.append(FlaskIntegration())

        try:
            if hasattr(config, 'SENTRY_TRANSPORT') and isinstance(config.SENTRY_TRANSPORT, tuple):
                mod = importlib.import_module(config.SENTRY_TRANSPORT[1])
                transport = getattr(mod, config.SENTRY_TRANSPORT[0])

                sentry_sdk.init(
                    dsn=config.SENTRY_DSN,
                    integrations=sentry_integrations,
                    transport=transport
                )
            else:
                sentry_sdk.init(
                    dsn=config.SENTRY_DSN,
                    integrations=sentry_integrations
                )
        except ImportError:
            log.exception(f'Unable to import selected SENTRY_TRANSPORT - {config.SENTRY_TRANSPORT}')
            exit(-1)

    logger.setLevel(config.BOT_LOG_LEVEL)

    storage_plugin = get_storage_plugin(config)

    # init the botplugin manager
    botplugins_dir = path.join(config.BOT_DATA_DIR, PLUGINS_SUBDIR)
    if not path.exists(botplugins_dir):
        makedirs(botplugins_dir, mode=0o755)

    plugin_indexes = getattr(config, 'BOT_PLUGIN_INDEXES', (PLUGIN_DEFAULT_INDEX,))
    if isinstance(plugin_indexes, str):
        plugin_indexes = (plugin_indexes, )

    # Extra backend is expected to be a list type, convert string to list.
    extra_backend = getattr(config, 'BOT_EXTRA_BACKEND_DIR', [])
    if isinstance(extra_backend, str):
        extra_backend = [extra_backend]

    backendpm = BackendPluginManager(config,
                                     'errbot.backends',
                                     backend_name,
                                     ErrBot,
                                     CORE_BACKENDS,
                                     extra_backend)

    log.info(f'Found Backend plugin: {backendpm.plugin_info.name}')

    repo_manager = BotRepoManager(storage_plugin,
                                  botplugins_dir,
                                  plugin_indexes)

    try:
        bot = backendpm.load_plugin()
        botpm = BotPluginManager(storage_plugin,
                                 config.BOT_EXTRA_PLUGIN_DIR,
                                 config.AUTOINSTALL_DEPS,
                                 getattr(config, 'CORE_PLUGINS', None),
                                 lambda name, clazz: clazz(bot, name),
                                 getattr(config, 'PLUGINS_CALLBACK_ORDER', (None, )))
        bot.attach_storage_plugin(storage_plugin)
        bot.attach_repo_manager(repo_manager)
        bot.attach_plugin_manager(botpm)
        bot.initialize_backend_storage()

        # restore the bot from the restore script
        if restore:
            # Prepare the context for the restore script
            if 'repos' in bot:
                log.fatal('You cannot restore onto a non empty bot.')
                sys.exit(-1)
            log.info(f'**** RESTORING the bot from {restore}')
            restore_bot_from_backup(restore, bot=bot, log=log)
            print('Restore complete. You can restart the bot normally')
            sys.exit(0)

        errors = bot.plugin_manager.update_plugin_places(repo_manager.get_all_repos_paths())
        if errors:
            log.error('Some plugins failed to load:\n' + '\n'.join(errors.values()))
            bot._plugin_errors_during_startup = "\n".join(errors.values())
        return bot
    except Exception:
        log.exception("Unable to load or configure the backend.")
        exit(-1)


def restore_bot_from_backup(backup_filename, *, bot, log):
    """Restores the given bot by executing the 'backup' script.

    The backup file is a python script which manually execute a series of commands on the bot
    to restore it to its previous state.

    :param backup_filename: the full path to the backup script.
    :param bot: the bot instance to restore
    :param log: logger to use during the restoration process
    """
    with open(backup_filename) as f:
        exec(f.read(), {'log': log, 'bot': bot})
    bot.close_storage()


def get_storage_plugin(config):
    """
    Find and load the storage plugin
    :param config: the bot configuration.
    :return: the storage plugin
    """
    storage_name = getattr(config, 'STORAGE', 'Shelf')
    extra_storage_plugins_dir = getattr(config, 'BOT_EXTRA_STORAGE_PLUGINS_DIR', None)
    spm = BackendPluginManager(config, 'errbot.storage', storage_name, StoragePluginBase,
                               CORE_STORAGE, extra_storage_plugins_dir)
    log.info(f'Found Storage plugin: {spm.plugin_info.name}.')
    return spm.load_plugin()


def bootstrap(bot_class, logger, config, restore=None):
    """
    Main starting point of Errbot.

    :param bot_class: The backend class inheriting from Errbot you want to start.
    :param logger: The logger you want to use.
    :param config: The config.py module.
    :param restore: Start Errbot in restore mode (from a backup).
    """
    bot = setup_bot(bot_class, logger, config, restore)
    log.debug(f'Start serving commands from the {bot.mode} backend.')
    bot.serve_forever()



import os
import errbot.backends.base
from errbot.backends.test import TestOccupant
import logging
log = logging.getLogger(__name__)

extra_plugin_dir = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'room_plugin')


def test_plugin_methods(testbot):
    p = testbot.bot.plugin_manager.get_plugin_obj_by_name('ChatRoom')
    assert p is not None

    assert hasattr(p, 'rooms')
    assert hasattr(p, 'query_room')


def test_create_join_leave_destroy_lifecycle(testbot):  # noqa
    rooms = testbot.bot.rooms()
    assert len(rooms) == 1

    r1 = rooms[0]
    assert str(r1) == "testroom"
    assert issubclass(r1.__class__, errbot.backends.base.Room)

    r2 = testbot.bot.query_room('testroom2')
    assert not r2.exists

    r2.create()
    assert r2.exists
    rooms = testbot.bot.rooms()
    assert r2 not in rooms
    assert not r2.joined

    r2.destroy()
    rooms = testbot.bot.rooms()
    assert r2 not in rooms

    r2.join()
    assert r2.exists
    assert r2.joined
    rooms = testbot.bot.rooms()
    assert r2 in rooms

    r2 = testbot.bot.query_room('testroom2')
    assert r2.joined

    r2.leave()
    assert not r2.joined
    r2.destroy()
    rooms = testbot.bot.rooms()
    assert r2 not in rooms


def test_occupants(testbot):  # noqa
    room = testbot.bot.rooms()[0]
    assert len(room.occupants) == 1
    assert TestOccupant('err', 'testroom') in room.occupants


def test_topic(testbot):  # noqa
    room = testbot.bot.rooms()[0]
    assert room.topic is None

    room.topic = "Errbot rocks!"
    assert room.topic == "Errbot rocks!"
    assert testbot.bot.rooms()[0].topic == "Errbot rocks!"


def test_plugin_callbacks(testbot):  # noqa
    p = testbot.bot.plugin_manager.get_plugin_obj_by_name('RoomTest')
    assert p is not None
    p.purge()

    log.debug("query and join")
    p.query_room('newroom').join()
    assert p.events.get(timeout=5) == "callback_room_joined newroom"

    p.query_room('newroom').topic = "Errbot rocks!"
    assert p.events.get(timeout=5) == "callback_room_topic Errbot rocks!"

    p.query_room('newroom').leave()
    assert p.events.get(timeout=5) == "callback_room_left newroom"


def test_botcommands(testbot):  # noqa
    rooms = testbot.bot.rooms()
    room = testbot.bot.query_room('testroom')
    assert len(rooms) == 1
    assert rooms[0] == room

    assert room.joined
    assert "Left the room testroom" in testbot.exec_command("!room leave testroom")
    room = testbot.bot.query_room('testroom')
    assert not room.joined

    assert "I'm not currently in any rooms." in testbot.exec_command("!room list")

    assert "Destroyed the room testroom" in testbot.exec_command("!room destroy testroom")

    rooms = testbot.bot.rooms()
    room = testbot.bot.query_room('testroom')
    assert not room.exists
    assert room not in rooms

    assert "Created the room testroom" in testbot.exec_command("!room create testroom")
    rooms = testbot.bot.rooms()
    room = testbot.bot.query_room('testroom')
    assert room.exists
    assert room not in rooms
    assert not room.joined

    assert "Joined the room testroom" in testbot.exec_command("!room join testroom")
    rooms = testbot.bot.rooms()
    room = testbot.bot.query_room('testroom')
    assert room.exists
    assert room.joined
    assert room in rooms

    assert "testroom" in testbot.exec_command("!room list")
    assert "err" in testbot.exec_command("!room occupants testroom")
    assert "No topic is set for testroom" in testbot.exec_command("!room topic testroom")
    assert "Topic for testroom set." in testbot.exec_command("!room topic testroom 'Errbot rocks!'")
    assert "Topic for testroom: Errbot rocks!" in testbot.exec_command("!room topic testroom")



import os

from errbot import BotPlugin, botcmd


class Backup(BotPlugin):
    """ Backup related commands. """
    @botcmd(admin_only=True)
    def backup(self, msg, args):
        """Backup everything.
           Makes a backup script called backup.py in the data bot directory.
           You can restore the backup from the command line with errbot --restore
        """
        filename = os.path.join(self.bot_config.BOT_DATA_DIR, 'backup.py')
        with open(filename, 'w') as f:
            f.write('## This file is not executable on its own. use errbot -r FILE to restore your bot.\n\n')
            f.write('log.info("Restoring repo_manager.")\n')
            for key, value in self._bot.repo_manager.items():
                f.write('bot.repo_manager["' + key + '"] = ' + repr(value) + '\n')
            f.write('log.info("Restoring plugin_manager.")\n')
            for key, value in self._bot.plugin_manager.items():  # don't mimic that in real plugins, this is core only.
                f.write('bot.plugin_manager["' + key + '"] = ' + repr(value) + '\n')

            f.write('log.info("Installing plugins.")\n')
            f.write('if "installed_repos" in bot.repo_manager:\n')
            f.write('  for repo in bot.repo_manager["installed_repos"]:\n')
            f.write('    log.error(bot.repo_manager.install_repo(repo))\n')

            f.write('log.info("Restoring plugins data.")\n')
            f.write('bot.plugin_manager.update_plugin_places(bot.repo_manager.get_all_repos_paths())\n')
            for plugin in self._bot.plugin_manager.plugins.values():
                if plugin._store:
                    f.write('pobj = bot.plugin_manager.plugins["' + plugin.name + '"]\n')
                    f.write('pobj.init_storage()\n')

                    for key, value in plugin.items():
                        f.write('pobj["' + key + '"] = ' + repr(value) + '\n')
                    f.write('pobj.close_storage()\n')

        return f'The backup file has been written in "{filename}".'



from __future__ import absolute_import
from errbot import BotPlugin, botcmd, Command, botmatch, arg_botcmd


def say_foo(plugin, msg, args):
    return 'foo %s' % type(plugin)


class Dyna(BotPlugin):
    """Just a test plugin to see if dynamic plugin API works.
    """
    @botcmd
    def add_simple(self, _, _1):
        simple1 = Command(lambda plugin, msg, args: 'yep %s' % type(plugin), name='say_yep')
        simple2 = Command(say_foo)

        self.create_dynamic_plugin('simple with special#', (simple1, simple2), doc='documented')

        return 'added'

    @botcmd
    def remove_simple(self, msg, args):
        self.destroy_dynamic_plugin('simple with special#')
        return 'removed'

    @botcmd
    def add_arg(self, _, _1):
        cmd1_name = 'echo_to_me'
        cmd1 = Command(lambda plugin, msg, args: 'string to echo is %s' % args.positional_arg,
                       cmd_type=arg_botcmd, cmd_args=('positional_arg',),
                       cmd_kwargs={'unpack_args': False, 'name': cmd1_name},
                       name=cmd1_name)

        self.create_dynamic_plugin('arg', (cmd1,), doc='documented')

        return 'added'

    @botcmd
    def remove_arg(self, msg, args):
        self.destroy_dynamic_plugin('arg')
        return 'removed'

    @botcmd
    def add_re(self, _, _1):
        re1 = Command(lambda plugin, msg, match: 'fffound',
                      name='ffound',
                      cmd_type=botmatch,
                      cmd_args=(r'^.*cheese.*$',))
        self.create_dynamic_plugin('re', (re1, ))
        return 'added'

    @botcmd
    def remove_re(self, msg, args):
        self.destroy_dynamic_plugin('re')
        return 'removed'

    @botcmd
    def add_saw(self, _, _1):
        re1 = Command(lambda plugin, msg, args: '+'.join(args),
                      name='splitme',
                      cmd_type=botcmd,
                      cmd_kwargs={'split_args_with': ','})
        self.create_dynamic_plugin('saw', (re1, ))
        return 'added'

    @botcmd
    def remove_saw(self, msg, args):
        self.destroy_dynamic_plugin('saw')
        return 'removed'

    @botcmd
    def clash(self, msg, args):
        return 'original'

    @botcmd
    def add_clashing(self, _, _1):
        simple1 = Command(lambda plugin, msg, args: 'dynamic', name='clash')
        self.create_dynamic_plugin('clashing', (simple1, ))
        return 'added'

    @botcmd
    def remove_clashing(self, _, _1):
        self.destroy_dynamic_plugin('clashing')
        return 'removed'



import logging
from errbot import BotPlugin
from errbot.core_plugins.webserver import webhook
from flask import abort, after_this_request

log = logging.getLogger(__name__)


class WebTest(BotPlugin):
    @webhook
    def webhook1(self, payload):
        log.debug(str(payload))
        return str(payload)

    @webhook(r'/custom_webhook')
    def webhook2(self, payload):
        log.debug(str(payload))
        return str(payload)

    @webhook(r'/form', form_param='form')
    def webhook3(self, payload):
        log.debug(str(payload))
        return str(payload)

    @webhook(r'/custom_form', form_param='form')
    def webhook4(self, payload):
        log.debug(str(payload))
        return str(payload)

    @webhook(r'/raw', raw=True)
    def webhook5(self, payload):
        log.debug(str(payload))
        return str(type(payload))

    @webhook
    def webhook6(self, payload):
        log.debug(str(payload))

        @after_this_request
        def add_header(response):
            response.headers['X-Powered-By'] = 'Errbot'
            return response
        return str(payload)

    @webhook
    def webhook7(self, payload):
        abort(403, "Forbidden")

    webhook8 = webhook(r'/lambda')(lambda x, y: str(x) + str(y))

    # Just to test https://github.com/errbotio/errbot/issues/1043
    @webhook(raw=True)
    def raw2(self, payload):
        log.debug(str(payload))
        return str(type(payload))



# -*- coding=utf-8 -*-
from os import path
# This is to test end2end i18n behavior.

extra_plugin_dir = path.join(path.dirname(path.realpath(__file__)), 'i18n_plugin')


def test_i18n_return(testbot):
    assert 'язы́к' in testbot.exec_command('!i18n 1')


def test_i18n_simple_name(testbot):
    assert 'OK' in testbot.exec_command('!ру́сский')


def test_i18n_prefix(testbot):
    assert 'OK' in testbot.exec_command('!prefix_ру́сский')
    assert 'OK' in testbot.exec_command('!prefix ру́сский')


def test_i18n_suffix(testbot):
    assert 'OK' in testbot.exec_command('!ру́сский_suffix')
    assert 'OK' in testbot.exec_command('!ру́сский suffix')
